# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper for using the Scikit-Learn API with Keras models."""


import copy
import types
import warnings

import numpy as np

from keras import losses
from keras.models import Sequential
from keras.utils.generic_utils import has_arg
from keras.utils.np_utils import to_categorical

# isort: off
from tensorflow.python.util.tf_export import keras_export
from tensorflow.tools.docs import doc_controls


class BaseWrapper:
    """Base class for the Keras scikit-learn wrapper.

    Warning: This class should not be used directly.
    Use descendant classes instead.

    Args:
        build_fn: callable function or class instance
        **sk_params: model parameters & fitting parameters

    The `build_fn` should construct, compile and return a Keras model, which
    will then be used to fit/predict. One of the following
    three values could be passed to `build_fn`:
    1. A function
    2. An instance of a class that implements the `__call__` method
    3. None. This means you implement a class that inherits from either
    `KerasClassifier` or `KerasRegressor`. The `__call__` method of the
    present class will then be treated as the default `build_fn`.

    `sk_params` takes both model parameters and fitting parameters. Legal model
    parameters are the arguments of `build_fn`. Note that like all other
    estimators in scikit-learn, `build_fn` should provide default values for
    its arguments, so that you could create the estimator without passing any
    values to `sk_params`.

    `sk_params` could also accept parameters for calling `fit`, `predict`,
    `predict_proba`, and `score` methods (e.g., `epochs`, `batch_size`).
    fitting (predicting) parameters are selected in the following order:

    1. Values passed to the dictionary arguments of
    `fit`, `predict`, `predict_proba`, and `score` methods
    2. Values passed to `sk_params`
    3. The default values of the `keras.models.Sequential`
    `fit`, `predict` methods.

    When using scikit-learn's `grid_search` API, legal tunable parameters are
    those you could pass to `sk_params`, including fitting parameters.
    In other words, you could use `grid_search` to search for the best
    `batch_size` or `epochs` as well as the model parameters.
    """

    def __init__(self, build_fn=None, **sk_params):
        self.build_fn = build_fn
        self.sk_params = sk_params
        self.check_params(sk_params)

    def check_params(self, params):
        """Checks for user typos in `params`.

        Args:
            params: dictionary; the parameters to be checked

        Raises:
            ValueError: if any member of `params` is not a valid argument.
        """
        legal_params_fns = [
            Sequential.fit,
            Sequential.predict,
            Sequential.evaluate,
        ]
        if self.build_fn is None:
            legal_params_fns.append(self.__call__)
        elif not isinstance(
            self.build_fn, types.FunctionType
        ) and not isinstance(self.build_fn, types.MethodType):
            legal_params_fns.append(self.build_fn.__call__)
        else:
            legal_params_fns.append(self.build_fn)

        for params_name in params:
            for fn in legal_params_fns:
                if has_arg(fn, params_name):
                    break
            else:
                if params_name != "nb_epoch":
                    raise ValueError(
                        "{} is not a legal parameter".format(params_name)
                    )

    def get_params(self, **params):
        """Gets parameters for this estimator.

        Args:
            **params: ignored (exists for API compatibility).

        Returns:
            Dictionary of parameter names mapped to their values.
        """
        res = self.sk_params.copy()
        res.update({"build_fn": self.build_fn})
        return res

    def set_params(self, **params):
        """Sets the parameters of this estimator.

        Args:
            **params: Dictionary of parameter names mapped to their values.

        Returns:
            self
        """
        self.check_params(params)
        self.sk_params.update(params)
        return self

    def fit(self, x, y, **kwargs):
        """Constructs a new model with `build_fn` & fit the model to `(x, y)`.

        Args:
            x : array-like, shape `(n_samples, n_features)`
                Training samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            y : array-like, shape `(n_samples,)` or `(n_samples, n_outputs)`
                True labels for `x`.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.fit`

        Returns:
            history : object
                details about the training history at each epoch.
        """
        if self.build_fn is None:
            self.model = self.__call__(**self.filter_sk_params(self.__call__))
        elif not isinstance(
            self.build_fn, types.FunctionType
        ) and not isinstance(self.build_fn, types.MethodType):
            self.model = self.build_fn(
                **self.filter_sk_params(self.build_fn.__call__)
            )
        else:
            self.model = self.build_fn(**self.filter_sk_params(self.build_fn))

        if (
            losses.is_categorical_crossentropy(self.model.loss)
            and len(y.shape) != 2
        ):
            y = to_categorical(y)

        fit_args = copy.deepcopy(self.filter_sk_params(Sequential.fit))
        fit_args.update(kwargs)

        history = self.model.fit(x, y, **fit_args)

        return history

    def filter_sk_params(self, fn, override=None):
        """Filters `sk_params` and returns those in `fn`'s arguments.

        Args:
            fn : arbitrary function
            override: dictionary, values to override `sk_params`

        Returns:
            res : dictionary containing variables
                in both `sk_params` and `fn`'s arguments.
        """
        override = override or {}
        res = {}
        for name, value in self.sk_params.items():
            if has_arg(fn, name):
                res.update({name: value})
        res.update(override)
        return res


@keras_export("keras.wrappers.scikit_learn.KerasClassifier")
@doc_controls.do_not_generate_docs
class KerasClassifier(BaseWrapper):
    """Implementation of the scikit-learn classifier API for Keras.

    DEPRECATED. Use [Sci-Keras](https://github.com/adriangb/scikeras) instead.
    See https://www.adriangb.com/scikeras/stable/migration.html
    for help migrating.
    """

    def __init__(self, build_fn=None, **sk_params):
        warnings.warn(
            "KerasClassifier is deprecated, "
            "use Sci-Keras (https://github.com/adriangb/scikeras) instead. "
            "See https://www.adriangb.com/scikeras/stable/migration.html "
            "for help migrating.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(build_fn, **sk_params)

    def fit(self, x, y, **kwargs):
        """Constructs a new model with `build_fn` & fit the model to `(x, y)`.

        Args:
            x : array-like, shape `(n_samples, n_features)`
                Training samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            y : array-like, shape `(n_samples,)` or `(n_samples, n_outputs)`
                True labels for `x`.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.fit`

        Returns:
            history : object
                details about the training history at each epoch.

        Raises:
            ValueError: In case of invalid shape for `y` argument.
        """
        y = np.array(y)
        if len(y.shape) == 2 and y.shape[1] > 1:
            self.classes_ = np.arange(y.shape[1])
        elif (len(y.shape) == 2 and y.shape[1] == 1) or len(y.shape) == 1:
            self.classes_ = np.unique(y)
            y = np.searchsorted(self.classes_, y)
        else:
            raise ValueError("Invalid shape for y: " + str(y.shape))
        self.n_classes_ = len(self.classes_)
        return super().fit(x, y, **kwargs)

    def predict(self, x, **kwargs):
        """Returns the class predictions for the given test data.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            **kwargs: dictionary arguments
                Legal arguments are the arguments
                of `Sequential.predict`.

        Returns:
            preds: array-like, shape `(n_samples,)`
                Class predictions.
        """
        proba = self.model.predict(x, **kwargs)
        if proba.shape[-1] > 1:
            classes = proba.argmax(axis=-1)
        else:
            classes = (proba > 0.5).astype("int32")
        return self.classes_[classes]

    def predict_proba(self, x, **kwargs):
        """Returns class probability estimates for the given test data.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            **kwargs: dictionary arguments
                Legal arguments are the arguments
                of `Sequential.predict`.

        Returns:
            proba: array-like, shape `(n_samples, n_outputs)`
                Class probability estimates.
                In the case of binary classification,
                to match the scikit-learn API,
                will return an array of shape `(n_samples, 2)`
                (instead of `(n_sample, 1)` as in Keras).
        """
        probs = self.model.predict(x, **kwargs)

        # check if binary classification
        if probs.shape[1] == 1:
            # first column is probability of class 0 and second is of class 1
            probs = np.hstack([1 - probs, probs])
        return probs

    def score(self, x, y, **kwargs):
        """Returns the mean accuracy on the given test data and labels.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            y: array-like, shape `(n_samples,)` or `(n_samples, n_outputs)`
                True labels for `x`.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.evaluate`.

        Returns:
            score: float
                Mean accuracy of predictions on `x` wrt. `y`.

        Raises:
            ValueError: If the underlying model isn't configured to
                compute accuracy. You should pass `metrics=["accuracy"]` to
                the `.compile()` method of the model.
        """
        y = np.searchsorted(self.classes_, y)
        kwargs = self.filter_sk_params(Sequential.evaluate, kwargs)

        loss_name = self.model.loss
        if hasattr(loss_name, "__name__"):
            loss_name = loss_name.__name__
        if loss_name == "categorical_crossentropy" and len(y.shape) != 2:
            y = to_categorical(y)

        outputs = self.model.evaluate(x, y, **kwargs)
        if not isinstance(outputs, list):
            outputs = [outputs]
        for name, output in zip(self.model.metrics_names, outputs):
            if name in ["accuracy", "acc"]:
                return output
        raise ValueError(
            "The model is not configured to compute accuracy. "
            'You should pass `metrics=["accuracy"]` to '
            "the `model.compile()` method."
        )


@keras_export("keras.wrappers.scikit_learn.KerasRegressor")
@doc_controls.do_not_generate_docs
class KerasRegressor(BaseWrapper):
    """Implementation of the scikit-learn regressor API for Keras.

    DEPRECATED. Use [Sci-Keras](https://github.com/adriangb/scikeras) instead.
    See https://www.adriangb.com/scikeras/stable/migration.html
    for help migrating.
    """

    @doc_controls.do_not_doc_inheritable
    def __init__(self, build_fn=None, **sk_params):
        warnings.warn(
            "KerasRegressor is deprecated, "
            "use Sci-Keras (https://github.com/adriangb/scikeras) instead. "
            "See https://www.adriangb.com/scikeras/stable/migration.html "
            "for help migrating.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(build_fn, **sk_params)

    def predict(self, x, **kwargs):
        """Returns predictions for the given test data.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.predict`.

        Returns:
            preds: array-like, shape `(n_samples,)`
                Predictions.
        """
        kwargs = self.filter_sk_params(Sequential.predict, kwargs)
        return np.squeeze(self.model.predict(x, **kwargs))

    def score(self, x, y, **kwargs):
        """Returns the mean loss on the given test data and labels.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            y: array-like, shape `(n_samples,)`
                True labels for `x`.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.evaluate`.

        Returns:
            score: float
                Mean accuracy of predictions on `x` wrt. `y`.
        """
        kwargs = self.filter_sk_params(Sequential.evaluate, kwargs)
        loss = self.model.evaluate(x, y, **kwargs)
        if isinstance(loss, list):
            return -loss[0]
        return -loss



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Constants for Keras SavedModel serialization."""

# Namespace used to store all attributes added during serialization.
# e.g. the list of layers can be accessed using `loaded.keras_api.layers`, in an
# object loaded from `tf.saved_model.load()`.
KERAS_ATTR = "keras_api"

# Keys for the serialization cache.
# Maps to the keras serialization dict {Layer --> SerializedAttributes object}
KERAS_CACHE_KEY = "keras_serialized_attributes"


# Name of Keras metadata file stored in the SavedModel.
SAVED_METADATA_PATH = "keras_metadata.pb"

# Names of SavedObject Keras identifiers.
INPUT_LAYER_IDENTIFIER = "_tf_keras_input_layer"
LAYER_IDENTIFIER = "_tf_keras_layer"
METRIC_IDENTIFIER = "_tf_keras_metric"
MODEL_IDENTIFIER = "_tf_keras_model"
NETWORK_IDENTIFIER = "_tf_keras_network"
RNN_LAYER_IDENTIFIER = "_tf_keras_rnn_layer"
SEQUENTIAL_IDENTIFIER = "_tf_keras_sequential"

KERAS_OBJECT_IDENTIFIERS = (
    INPUT_LAYER_IDENTIFIER,
    LAYER_IDENTIFIER,
    METRIC_IDENTIFIER,
    MODEL_IDENTIFIER,
    NETWORK_IDENTIFIER,
    RNN_LAYER_IDENTIFIER,
    SEQUENTIAL_IDENTIFIER,
)



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tf.keras models with callbacks, checkpointing with dist
strategy."""

import collections
import tempfile

import numpy as np
import tensorflow.compat.v2 as tf
from absl.testing import parameterized

import keras
from keras import losses
from keras.distribute import distribute_strategy_test as keras_test_lib
from keras.distribute import distributed_training_utils_v1
from keras.distribute import optimizer_combinations


class Counter(keras.callbacks.Callback):
    """Counts the number of times each callback method was run.

    Attributes:
      method_counts: dict. Contains the counts of time  each callback method was
        run.
    """

    def __init__(self):
        self.method_counts = collections.defaultdict(int)
        methods_to_count = [
            "on_batch_begin",
            "on_batch_end",
            "on_epoch_begin",
            "on_epoch_end",
            "on_predict_batch_begin",
            "on_predict_batch_end",
            "on_predict_begin",
            "on_predict_end",
            "on_test_batch_begin",
            "on_test_batch_end",
            "on_test_begin",
            "on_test_end",
            "on_train_batch_begin",
            "on_train_batch_end",
            "on_train_begin",
            "on_train_end",
        ]
        for method_name in methods_to_count:
            setattr(
                self,
                method_name,
                self.wrap_with_counts(method_name, getattr(self, method_name)),
            )

    def wrap_with_counts(self, method_name, method):
        def _call_and_count(*args, **kwargs):
            self.method_counts[method_name] += 1
            return method(*args, **kwargs)

        return _call_and_count


class TestDistributionStrategyWithCallbacks(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations()
        )
    )
    def test_callbacks_in_fit(self, distribution):
        with distribution.scope():
            model = keras_test_lib.get_model()
            model.compile(optimizer="sgd", loss="mse", metrics=["mae"])

        dataset = keras_test_lib.get_dataset(distribution)
        counter = Counter()

        epochs = 2
        steps_per_epoch = 5
        validation_steps = 3

        model.fit(
            dataset,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            verbose=0,
            validation_data=dataset,
            validation_steps=validation_steps,
            callbacks=[counter],
        )

        if (
            isinstance(
                distribution, tf.compat.v1.distribute.experimental.TPUStrategy
            )
            and not tf.executing_eagerly()
        ):
            # TPU Strategy can have multi step training, from
            # extended.steps_per_run if steps_per_run = 1, then
            # num_batch_call_per_epoch = steps_per_epoch
            steps_per_run = distribution.extended.steps_per_run
            num_batch_call_per_epoch = steps_per_epoch // steps_per_run
            if steps_per_epoch % steps_per_run:
                num_batch_call_per_epoch += 1
        else:
            num_batch_call_per_epoch = steps_per_epoch

        self.assertDictEqual(
            counter.method_counts,
            {
                "on_batch_begin": epochs * num_batch_call_per_epoch,
                "on_batch_end": epochs * num_batch_call_per_epoch,
                "on_epoch_begin": epochs,
                "on_epoch_end": epochs,
                "on_test_batch_begin": epochs * validation_steps,
                "on_test_batch_end": epochs * validation_steps,
                "on_test_begin": epochs,
                "on_test_end": epochs,
                "on_train_batch_begin": epochs * num_batch_call_per_epoch,
                "on_train_batch_end": epochs * num_batch_call_per_epoch,
                "on_train_begin": 1,
                "on_train_end": 1,
            },
        )

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations()
        )
    )
    def test_callbacks_in_eval(self, distribution):
        with distribution.scope():
            model = keras_test_lib.get_model()
            model.compile(optimizer="sgd", loss="mse", metrics=["mae"])

        dataset = keras_test_lib.get_dataset(distribution)
        counter = Counter()

        model.evaluate(dataset, steps=5, callbacks=[counter])

        self.assertDictEqual(
            counter.method_counts,
            {
                "on_test_batch_begin": 5,
                "on_test_batch_end": 5,
                "on_test_begin": 1,
                "on_test_end": 1,
            },
        )

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations()
        )
    )
    def test_callbacks_in_predict(self, distribution):
        with distribution.scope():
            model = keras_test_lib.get_model()
            model.compile(optimizer="sgd", loss="mse", metrics=["mae"])

        dataset = keras_test_lib.get_dataset(distribution)
        counter = Counter()

        model.predict(
            keras_test_lib.get_predict_dataset(dataset),
            steps=5,
            callbacks=[counter],
        )

        self.assertDictEqual(
            counter.method_counts,
            {
                "on_predict_batch_begin": 5,
                "on_predict_batch_end": 5,
                "on_predict_begin": 1,
                "on_predict_end": 1,
            },
        )


class TestDistributionStrategyErrorCases(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph"],
        )
    )
    def test_validating_dataset_input_tensors_with_shape_mismatch(
        self, distribution
    ):
        with self.cached_session():

            @tf.function
            def run():
                ctx = tf.distribute.get_replica_context()
                if ctx.replica_id_in_sync_group.device.endswith("GPU:0"):
                    return tf.constant([[1, 2]])
                else:
                    return tf.constant([[1, 2], [1, 2]])

            x = distribution.run(run)

            # Removed device and input tensor shape details from the error
            # message since the order of the device and the corresponding input
            # tensor shape is not deterministic over different runs.
            with self.assertRaisesRegex(
                ValueError,
                "Input tensor shapes do not match for "
                "distributed tensor inputs "
                "PerReplica:.+",
            ):
                with distribution.scope():
                    distributed_training_utils_v1.validate_distributed_dataset_inputs(  # noqa: E501
                        distribution, x, None
                    )

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph", "eager"],
        )
    )
    def test_validating_dataset_input_tensors_with_dtype_mismatch(
        self, distribution
    ):
        with self.cached_session():

            @tf.function
            def run():
                ctx = tf.distribute.get_replica_context()
                if ctx.replica_id_in_sync_group.device.endswith("GPU:0"):
                    return tf.constant([[1, 2]], dtype=tf.int32)
                else:
                    return tf.constant([[1, 2]], dtype=tf.float64)

            x = distribution.run(run)

            # Removed device and input tensor dtype details from the error
            # message since the order of the device and the corresponding input
            # tensor dtype is not deterministic over different runs.
            with self.assertRaisesRegex(
                ValueError,
                "Input tensor dtypes do not match for "
                "distributed tensor inputs "
                "PerReplica:.+",
            ):
                with distribution.scope():
                    distributed_training_utils_v1.validate_distributed_dataset_inputs(  # noqa: E501
                        distribution, x, None
                    )

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph", "eager"],
        )
    )
    def test_unsupported_features(self, distribution, mode):
        with self.cached_session():
            with distribution.scope():
                model = keras_test_lib.get_model()
                optimizer = tf.compat.v1.train.GradientDescentOptimizer(0.001)
                loss = "mse"
                metrics = ["mae"]
                model.compile(optimizer, loss, metrics=metrics)

            dataset = keras_test_lib.get_dataset(distribution)
            # Test with validation split
            with self.assertRaises(ValueError):
                model.fit(
                    dataset,
                    epochs=1,
                    steps_per_epoch=2,
                    verbose=0,
                    validation_split=0.5,
                    validation_steps=2,
                )

            # Test with sample weight.
            sample_weight = np.random.random((10,))
            with self.assertRaises(ValueError):
                model.fit(
                    dataset,
                    epochs=1,
                    steps_per_epoch=2,
                    verbose=0,
                    sample_weight=sample_weight,
                )

            # Test with not specifying the `steps` argument for dataset with
            # infinite cardinality.
            dataset = dataset.repeat()
            with self.assertRaises(ValueError):
                model.fit(dataset, epochs=1, verbose=0)
            with self.assertRaises(ValueError):
                model.evaluate(dataset, verbose=0)

            with self.assertRaises(ValueError):
                model.predict(dataset, verbose=0)

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
                tf.__internal__.distribute.combinations.one_device_strategy,
            ],
            mode=["graph", "eager"],
        )
    )
    def test_distribution_strategy_on_subclassed_model(self, distribution):
        with distribution.scope():

            class _SimpleMLP(keras.Model):
                def __init__(self, num_labels):
                    super().__init__()
                    self.dense = keras.layers.Dense(num_labels)

                def call(self, inputs):
                    return self.dense(inputs)

            model = _SimpleMLP(3)

            if not tf.executing_eagerly():
                with self.assertRaisesRegex(
                    ValueError,
                    "We currently do not support distribution strategy with a "
                    "`Sequential` model that is created without `input_shape`/"
                    "`input_dim` set in its first layer or a subclassed model.",
                ):
                    model.compile("sgd")
            else:
                model.compile("sgd")

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
                tf.__internal__.distribute.combinations.one_device_strategy,
            ],
            mode=["graph", "eager"],
        )
    )
    def test_distribution_strategy_on_deferred_sequential_model(
        self, distribution
    ):
        with distribution.scope():
            model = keras.models.Sequential()
            model.add(keras.layers.Dense(16, activation="relu"))
            model.add(keras.layers.Dense(3, activation="softmax"))

            if tf.executing_eagerly():
                model.compile("sgd")
            else:
                with self.assertRaisesRegex(
                    ValueError,
                    "We currently do not support distribution strategy with a "
                    "`Sequential` model that is created without "
                    "`input_shape`/`input_dim` set in its first layer or "
                    "a subclassed model.",
                ):
                    model.compile("sgd")

    @tf.__internal__.distribute.combinations.generate(
        keras_test_lib.all_strategy_combinations_minus_default()
    )
    def test_standalone_loss_without_loss_reduction(self, distribution):
        with distribution.scope():
            loss_object = losses.MeanSquaredError()

            with self.assertRaisesRegex(
                ValueError,
                "Please use `tf.keras.losses.Reduction.SUM` or "
                "`tf.keras.losses.Reduction.NONE`",
            ):
                y = np.asarray([1, 0])
                loss_object(y, y)


class TestDistributionStrategyWithLossMasking(
    tf.test.TestCase, parameterized.TestCase
):

    # TODO(priyag): Enable all strategies for this test. Currently it does not
    # work for TPU due to some invalid datatype.
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph", "eager"],
            optimizer=optimizer_combinations.gradient_descent_optimizer_keras_v2_fn,  # noqa: E501
        )
    )
    def test_masking(self, distribution, optimizer):
        with self.cached_session():
            np.random.seed(1337)
            x = np.array([[[1], [1]], [[0], [0]]])
            with distribution.scope():
                model = keras.models.Sequential()
                model.add(
                    keras.layers.Masking(mask_value=0, input_shape=(2, 1))
                )
                model.add(
                    keras.layers.TimeDistributed(
                        keras.layers.Dense(1, kernel_initializer="one")
                    )
                )
                model.compile(loss="mse", optimizer=optimizer())
            y = np.array([[[1], [1]], [[1], [1]]])
            dataset = tf.data.Dataset.from_tensor_slices((x, y))
            dataset = dataset.repeat(100)
            dataset = dataset.batch(10)
            hist = model.fit(x=dataset, epochs=1, steps_per_epoch=2)
            self.assertEqual(hist.history["loss"][0], 0)


class TestDistributionStrategyWithNormalizationLayer(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations(),
            tf.__internal__.test.combinations.combine(
                fused=[True, False],
                optimizer=optimizer_combinations.gradient_descent_optimizer_keras_v2_fn,  # noqa: E501
            ),
        )
    )
    def test_batchnorm_correctness(self, distribution, fused, optimizer):
        with self.cached_session():
            with distribution.scope():
                model = keras.models.Sequential()
                norm = keras.layers.BatchNormalization(
                    input_shape=(
                        10,
                        20,
                        30,
                    ),
                    momentum=0.8,
                    fused=fused,
                )
                model.add(norm)
                model.compile(loss="mse", optimizer=optimizer())

            # centered on 5.0, variance 10.0
            x = np.random.normal(loc=5.0, scale=10.0, size=(1000, 10, 20, 30))
            x = x.astype("float32")
            dataset = tf.data.Dataset.from_tensor_slices((x, x))
            dataset = dataset.repeat(100)
            dataset = keras_test_lib.batch_wrapper(dataset, 32, distribution)

            predict_dataset = tf.data.Dataset.from_tensor_slices(x)
            predict_dataset = predict_dataset.repeat(100)
            predict_dataset = keras_test_lib.batch_wrapper(
                predict_dataset, 32, distribution
            )

            model.fit(dataset, epochs=4, verbose=0, steps_per_epoch=10)
            out = model.predict(predict_dataset, steps=2)
            out -= keras.backend.eval(norm.beta)
            out /= keras.backend.eval(norm.gamma)
            np.testing.assert_allclose(out.mean(), 0.0, atol=1e-1)
            np.testing.assert_allclose(out.std(), 1.0, atol=1e-1)

    # TODO(b/146181571): Enable this for all distribution strategies once
    # DistributedVariable.assign() returns a variable for MirroredStrategy.
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.tpu_strategy_combinations(),
            tf.__internal__.test.combinations.combine(
                optimizer=optimizer_combinations.gradient_descent_optimizer_keras_v2_fn  # noqa: E501
            ),
        )
    )
    def test_batchnorm_correctness_with_renorm(self, distribution, optimizer):
        with self.cached_session():
            with distribution.scope():
                model = keras.models.Sequential()
                norm = keras.layers.BatchNormalization(
                    input_shape=(
                        10,
                        20,
                        30,
                    ),
                    momentum=0.8,
                    fused=False,
                    renorm=True,
                )
                model.add(norm)
                model.compile(loss="mse", optimizer=optimizer())

            # centered on 5.0, variance 10.0
            x = np.random.normal(loc=5.0, scale=10.0, size=(1000, 10, 20, 30))
            x = x.astype("float32")
            dataset = tf.data.Dataset.from_tensor_slices((x, x))
            dataset = dataset.repeat(100)
            dataset = keras_test_lib.batch_wrapper(dataset, 32, distribution)

            predict_dataset = tf.data.Dataset.from_tensor_slices(x)
            predict_dataset = predict_dataset.repeat(100)
            predict_dataset = keras_test_lib.batch_wrapper(
                predict_dataset, 32, distribution
            )

            model.fit(dataset, epochs=4, verbose=0, steps_per_epoch=10)
            out = model.predict(predict_dataset, steps=2)
            out -= keras.backend.eval(norm.beta)
            out /= keras.backend.eval(norm.gamma)
            np.testing.assert_allclose(out.mean(), 0.0, atol=1e-1)
            np.testing.assert_allclose(out.std(), 1.0, atol=1e-1)


class TestDistributionStrategySaveLoadWeights(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations_minus_default(),
            tf.__internal__.test.combinations.combine(
                optimizer=optimizer_combinations.rmsprop_optimizer_keras_v2_fn
            ),
        )
    )
    def test_save_load_h5(self, distribution, optimizer):
        with self.cached_session():
            dataset = keras_test_lib.get_dataset(distribution)
            with distribution.scope():
                model = keras_test_lib.get_model()
                model.compile(optimizer(), "mse")
                model.fit(dataset, epochs=1, steps_per_epoch=1)

                weights_file = tempfile.mktemp(".h5")
                model.save_weights(weights_file)

                model_2 = keras_test_lib.get_model()
                model_2.compile(optimizer(), "mse")
                model_2.load_weights(weights_file)
                model_2.predict(
                    keras_test_lib.get_predict_dataset(distribution), steps=2
                )
                model_2.fit(dataset, epochs=1, steps_per_epoch=1)

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations_minus_default(),
            tf.__internal__.test.combinations.combine(
                optimizer=optimizer_combinations.rmsprop_optimizer_keras_v2_fn
            ),
        )
    )
    def test_save_load_trackable(self, distribution, optimizer):
        # TODO(b/123533246): Enable the test for TPU once bug is fixed
        if (
            isinstance(
                distribution,
                (
                    tf.distribute.experimental.TPUStrategy,
                    tf.compat.v1.distribute.experimental.TPUStrategy,
                ),
            )
            and distribution.extended.steps_per_run > 1
        ):
            self.skipTest(
                "MultiStep TPU Strategy deadlocks with optimizer restore."
            )
        with self.cached_session():
            dataset = keras_test_lib.get_dataset(distribution)
            with distribution.scope():
                model = keras_test_lib.get_model()
                model.compile(optimizer(), "mse")
                model.fit(dataset, epochs=1, steps_per_epoch=1)

                weights_file = tempfile.mktemp()
                model.save_weights(weights_file)

                model_2 = keras_test_lib.get_model()
                model_2.compile(optimizer(), "mse")
                model_2.load_weights(weights_file)
                model_2.predict(
                    keras_test_lib.get_predict_dataset(distribution), steps=2
                )
                model_2.fit(dataset, epochs=1, steps_per_epoch=1)


class TestDistributionStrategyValidation(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_test_lib.all_strategy_combinations_minus_default()
        )
    )
    def test_layer_outside_scope(self, distribution):
        with self.cached_session():
            with self.assertRaisesRegex(
                ValueError, "was not created in the distribution strategy"
            ):
                x = keras.layers.Input(shape=(3,), name="input")
                y = keras.layers.Dense(4, name="dense")(x)
                with distribution.scope():
                    model = keras.Model(x, y)
                    optimizer = tf.compat.v1.train.GradientDescentOptimizer(
                        0.001
                    )
                    loss = "mse"
                    metrics = ["mae", keras.metrics.CategoricalAccuracy()]
                    model.compile(optimizer, loss, metrics=metrics)

    @tf.__internal__.distribute.combinations.generate(
        keras_test_lib.all_strategy_combinations_minus_default()
    )
    def test_model_outside_scope(self, distribution):
        with self.cached_session():
            with self.assertRaisesRegex(
                ValueError, "was not created in the distribution strategy"
            ):
                x = keras.layers.Input(shape=(3,), name="input")
                y = keras.layers.Dense(4, name="dense")(x)
                model = keras.Model(x, y)
                with distribution.scope():
                    optimizer = tf.compat.v1.train.GradientDescentOptimizer(
                        0.001
                    )
                    loss = "mse"
                    metrics = ["mae", keras.metrics.CategoricalAccuracy()]
                    model.compile(optimizer, loss, metrics=metrics)


class TestDistributionStrategyWithStaticShapes(
    tf.test.TestCase, parameterized.TestCase
):
    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph", "eager"],
        )
    )
    def test_input_batch_size_not_divisible_by_num_replicas(self, distribution):
        with distribution.scope():
            with self.assertRaisesRegex(
                ValueError,
                r"The `batch_size` argument \(5\) must be divisible by "
                r"the number of replicas \(2\)",
            ):
                keras.layers.Input(shape=(3,), batch_size=5, name="input")

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.combine(
            distribution=[
                tf.__internal__.distribute.combinations.mirrored_strategy_with_gpu_and_cpu,  # noqa: E501
            ],
            mode=["graph", "eager"],
        )
    )
    def test_static_input_batch_size(self, distribution):
        inputs = np.zeros((10, 3), dtype=np.float32)
        targets = np.zeros((10, 4), dtype=np.float32)
        dataset = tf.data.Dataset.from_tensor_slices((inputs, targets))
        dataset = dataset.repeat(100)
        dataset = dataset.batch(10, drop_remainder=True)

        with distribution.scope():
            x = keras.layers.Input(shape=(3,), batch_size=10, name="input")
            y = keras.layers.Dense(4, name="dense")(x)
            model = keras.Model(x, y)
            model.compile(optimizer="sgd", loss="mse", metrics=["mae"])

        model.fit(dataset, epochs=1, steps_per_epoch=5)
        model.evaluate(dataset, steps=5)
        model.predict(dataset)


if __name__ == "__main__":
    tf.__internal__.distribute.multi_process_runner.test_main()



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""ResNet-RS models for Keras.

Reference:
- [Revisiting ResNets: Improved Training and Scaling Strategies](
    https://arxiv.org/pdf/2103.07579.pdf)

"""

import sys as _sys

from keras.applications.resnet_rs import ResNetRS101
from keras.applications.resnet_rs import ResNetRS152
from keras.applications.resnet_rs import ResNetRS200
from keras.applications.resnet_rs import ResNetRS270
from keras.applications.resnet_rs import ResNetRS350
from keras.applications.resnet_rs import ResNetRS420
from keras.applications.resnet_rs import ResNetRS50
from keras.applications.resnet_rs import decode_predictions
from keras.applications.resnet_rs import preprocess_input


# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Constraints: functions that impose constraints on weight values.
"""

import sys as _sys

from keras.constraints import Constraint
from keras.constraints import MaxNorm
from keras.constraints import MaxNorm as max_norm
from keras.constraints import MinMaxNorm
from keras.constraints import MinMaxNorm as min_max_norm
from keras.constraints import NonNeg
from keras.constraints import NonNeg as non_neg
from keras.constraints import RadialConstraint
from keras.constraints import RadialConstraint as radial_constraint
from keras.constraints import UnitNorm
from keras.constraints import UnitNorm as unit_norm
from keras.constraints import deserialize
from keras.constraints import get
from keras.constraints import serialize


# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Fast GRU layer backed by cuDNN."""


import collections

import tensorflow.compat.v2 as tf

from keras import constraints
from keras import initializers
from keras import regularizers
from keras.layers.rnn import gru_lstm_utils
from keras.layers.rnn.base_cudnn_rnn import _CuDNNRNN

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export(v1=["keras.layers.CuDNNGRU"])
class CuDNNGRU(_CuDNNRNN):
    """Fast GRU implementation backed by cuDNN.

    More information about cuDNN can be found on the [NVIDIA
    developer website](https://developer.nvidia.com/cudnn).
    Can only be run on GPU.

    Args:
        units: Positive integer, dimensionality of the output space.
        kernel_initializer: Initializer for the `kernel` weights matrix, used
          for the linear transformation of the inputs.
        recurrent_initializer: Initializer for the `recurrent_kernel` weights
          matrix, used for the linear transformation of the recurrent state.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to the `kernel` weights
          matrix.
        recurrent_regularizer: Regularizer function applied to the
          `recurrent_kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to the output of the
          layer (its "activation").
        kernel_constraint: Constraint function applied to the `kernel` weights
          matrix.
        recurrent_constraint: Constraint function applied to the
          `recurrent_kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        return_sequences: Boolean. Whether to return the last output in the
          output sequence, or the full sequence.
        return_state: Boolean. Whether to return the last state in addition to
          the output.
        go_backwards: Boolean (default False). If True, process the input
          sequence backwards and return the reversed sequence.
        stateful: Boolean (default False). If True, the last state for each
          sample at index i in a batch will be used as initial state for the
          sample of index i in the following batch.
    """

    def __init__(
        self,
        units,
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        **kwargs
    ):
        self.units = units
        cell_spec = collections.namedtuple("cell", "state_size")
        self._cell = cell_spec(state_size=self.units)
        super().__init__(
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            **kwargs
        )

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.recurrent_constraint = constraints.get(recurrent_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

    @property
    def cell(self):
        return self._cell

    def build(self, input_shape):
        super().build(input_shape)
        if isinstance(input_shape, list):
            input_shape = input_shape[0]
        input_dim = int(input_shape[-1])

        self.kernel = self.add_weight(
            shape=(input_dim, self.units * 3),
            name="kernel",
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
        )

        self.recurrent_kernel = self.add_weight(
            shape=(self.units, self.units * 3),
            name="recurrent_kernel",
            initializer=self.recurrent_initializer,
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint,
        )

        self.bias = self.add_weight(
            shape=(self.units * 6,),
            name="bias",
            initializer=self.bias_initializer,
            regularizer=self.bias_regularizer,
            constraint=self.bias_constraint,
        )

        self.built = True

    def _process_batch(self, inputs, initial_state):
        if not self.time_major:
            inputs = tf.transpose(inputs, perm=(1, 0, 2))
        input_h = initial_state[0]
        input_h = tf.expand_dims(input_h, axis=0)

        params = gru_lstm_utils.canonical_to_params(
            weights=[
                self.kernel[:, self.units : self.units * 2],
                self.kernel[:, : self.units],
                self.kernel[:, self.units * 2 :],
                self.recurrent_kernel[:, self.units : self.units * 2],
                self.recurrent_kernel[:, : self.units],
                self.recurrent_kernel[:, self.units * 2 :],
            ],
            biases=[
                self.bias[self.units : self.units * 2],
                self.bias[: self.units],
                self.bias[self.units * 2 : self.units * 3],
                self.bias[self.units * 4 : self.units * 5],
                self.bias[self.units * 3 : self.units * 4],
                self.bias[self.units * 5 :],
            ],
            shape=self._vector_shape,
        )

        args = {
            "input": inputs,
            "input_h": input_h,
            "input_c": 0,
            "params": params,
            "is_training": True,
            "rnn_mode": "gru",
        }

        outputs, h, _, _, _ = tf.raw_ops.CudnnRNNV2(**args)

        if self.stateful or self.return_state:
            h = h[0]
        if self.return_sequences:
            if self.time_major:
                output = outputs
            else:
                output = tf.transpose(outputs, perm=(1, 0, 2))
        else:
            output = outputs[-1]
        return output, [h]

    def get_config(self):
        config = {
            "units": self.units,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "recurrent_initializer": initializers.serialize(
                self.recurrent_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "recurrent_regularizer": regularizers.serialize(
                self.recurrent_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "activity_regularizer": regularizers.serialize(
                self.activity_regularizer
            ),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(
                self.recurrent_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Layer that multiplies (element-wise) several inputs."""


from keras.layers.merging.base_merge import _Merge

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.Multiply")
class Multiply(_Merge):
    """Layer that multiplies (element-wise) a list of inputs.

    It takes as input a list of tensors, all of the same shape, and returns
    a single tensor (also of the same shape).

    >>> tf.keras.layers.Multiply()([np.arange(5).reshape(5, 1),
    ...                             np.arange(5, 10).reshape(5, 1)])
    <tf.Tensor: shape=(5, 1), dtype=int64, numpy=
    array([[ 0],
         [ 6],
         [14],
         [24],
         [36]])>

    >>> x1 = tf.keras.layers.Dense(8)(np.arange(10).reshape(5, 2))
    >>> x2 = tf.keras.layers.Dense(8)(np.arange(10, 20).reshape(5, 2))
    >>> multiplied = tf.keras.layers.Multiply()([x1, x2])
    >>> multiplied.shape
    TensorShape([5, 8])
    """

    def _merge_function(self, inputs):
        output = inputs[0]
        for i in range(1, len(inputs)):
            output = output * inputs[i]
        return output


@keras_export("keras.layers.multiply")
def multiply(inputs, **kwargs):
    """Functional interface to the `Multiply` layer.

    Example:

    >>> x1 = np.arange(3.0)
    >>> x2 = np.arange(3.0)
    >>> tf.keras.layers.multiply([x1, x2])
    <tf.Tensor: shape=(3,), dtype=float32, numpy=array([0., 1., 4.], ...)>

    Usage in a functional model:

    >>> input1 = tf.keras.layers.Input(shape=(16,))
    >>> x1 = tf.keras.layers.Dense(
    ...     8, activation='relu')(input1) #shape=(None, 8)
    >>> input2 = tf.keras.layers.Input(shape=(32,))
    >>> x2 = tf.keras.layers.Dense(
    ...     8, activation='relu')(input2) #shape=(None, 8)
    >>> out = tf.keras.layers.multiply([x1,x2]) #shape=(None, 8)
    >>> out = tf.keras.layers.Dense(4)(out)
    >>> model = tf.keras.models.Model(inputs=[input1, input2], outputs=out)

    Args:
        inputs: A list of input tensors.
        **kwargs: Standard layer keyword arguments.

    Returns:
        A tensor, the element-wise product of the inputs.
    """
    return Multiply(**kwargs)(inputs)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Legacy Adam optimizer implementation."""

from keras.optimizers.optimizer_v2 import optimizer_v2

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.optimizers.legacy.Optimizer")
class Optimizer(optimizer_v2.OptimizerV2):
    pass



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Test related utilities for KPL + tf.distribute."""

import random
import tempfile

import tensorflow.compat.v2 as tf

import keras
from keras.layers.preprocessing import string_lookup


class DistributeKplTestUtils(tf.test.TestCase):
    """Utils for test of tf.distribute + KPL."""

    FEATURE_VOCAB = [
        "avenger",
        "ironman",
        "batman",
        "hulk",
        "spiderman",
        "kingkong",
        "wonder_woman",
    ]
    LABEL_VOCAB = ["yes", "no"]

    def define_kpls_for_training(self, use_adapt):
        """Function that defines KPL used for unit tests of tf.distribute.

        Args:
          use_adapt: if adapt will be called. False means there will be
            precomputed statistics.

        Returns:
          feature_mapper: a simple keras model with one keras StringLookup layer
          which maps feature to index.
          label_mapper: similar to feature_mapper, but maps label to index.

        """
        if use_adapt:
            feature_lookup_layer = string_lookup.StringLookup(num_oov_indices=1)
            feature_lookup_layer.adapt(self.FEATURE_VOCAB)
            label_lookup_layer = string_lookup.StringLookup(
                num_oov_indices=0, mask_token=None
            )
            label_lookup_layer.adapt(self.LABEL_VOCAB)
        else:
            feature_lookup_layer = string_lookup.StringLookup(
                vocabulary=self.FEATURE_VOCAB, num_oov_indices=1
            )
            label_lookup_layer = string_lookup.StringLookup(
                vocabulary=self.LABEL_VOCAB, num_oov_indices=0, mask_token=None
            )

        raw_feature_input = keras.layers.Input(
            shape=(3,), dtype=tf.string, name="feature", ragged=True
        )
        feature_id_input = feature_lookup_layer(raw_feature_input)
        feature_mapper = keras.Model(
            {"features": raw_feature_input}, feature_id_input
        )

        raw_label_input = keras.layers.Input(
            shape=(1,), dtype=tf.string, name="label"
        )
        label_id_input = label_lookup_layer(raw_label_input)
        label_mapper = keras.Model({"label": raw_label_input}, label_id_input)

        return feature_mapper, label_mapper

    def dataset_fn(self, feature_mapper, label_mapper):
        """Function that generates dataset for test of tf.distribute + KPL.

        Args:
          feature_mapper: a simple keras model with one keras StringLookup layer
            which maps feature to index.
          label_mapper: similar to feature_mapper, but maps label to index.

        Returns:
          Generated dataset for test of tf.distribute + KPL.

        """

        def feature_and_label_gen():
            # Generator of dataset.
            while True:
                features = random.sample(self.FEATURE_VOCAB, 3)
                label = ["yes"] if self.FEATURE_VOCAB[0] in features else ["no"]
                yield {"features": features, "label": label}

        raw_dataset = (
            tf.data.Dataset.from_generator(
                feature_and_label_gen,
                output_signature={
                    "features": tf.TensorSpec([3], tf.string),
                    "label": tf.TensorSpec([1], tf.string),
                },
            )
            .shuffle(100)
            .batch(32)
        )

        train_dataset = raw_dataset.map(
            lambda x: (
                {"features": feature_mapper(x["features"])},
                label_mapper(x["label"]),
            )
        )
        return train_dataset

    def define_model(self):
        """A simple model for test of tf.distribute + KPL."""
        # Create the model. The input needs to be compatible with KPLs.
        model_input = keras.layers.Input(
            shape=(3,), dtype=tf.int64, name="model_input"
        )

        # input_dim includes a mask token and an oov token.
        emb_output = keras.layers.Embedding(
            input_dim=len(self.FEATURE_VOCAB) + 2, output_dim=20
        )(model_input)
        emb_output = tf.reduce_mean(emb_output, axis=1)
        dense_output = keras.layers.Dense(units=1, activation="sigmoid")(
            emb_output
        )
        model = keras.Model({"features": model_input}, dense_output)
        return model

    def define_reverse_lookup_layer(self):
        """Create string reverse lookup layer for serving."""

        label_inverse_lookup_layer = string_lookup.StringLookup(
            num_oov_indices=0,
            mask_token=None,
            vocabulary=self.LABEL_VOCAB,
            invert=True,
        )
        return label_inverse_lookup_layer

    def create_serving_signature(
        self, model, feature_mapper, label_inverse_lookup_layer
    ):
        """Create serving signature for the given model."""

        @tf.function
        def serve_fn(raw_features):
            raw_features = tf.expand_dims(raw_features, axis=0)
            transformed_features = model.feature_mapper(raw_features)
            outputs = model(transformed_features)
            outputs = tf.squeeze(outputs, axis=0)
            outputs = tf.cast(tf.greater(outputs, 0.5), tf.int64)
            decoded_outputs = model.label_inverse_lookup_layer(outputs)
            return tf.squeeze(decoded_outputs, axis=0)

        model.feature_mapper = feature_mapper
        model.label_inverse_lookup_layer = label_inverse_lookup_layer
        # serving does NOT have batch dimension
        return serve_fn.get_concrete_function(
            tf.TensorSpec(shape=(3), dtype=tf.string, name="example")
        )

    def test_save_load_serving_model(
        self, model, feature_mapper, label_inverse_lookup_layer
    ):
        """Test save/load/serving model."""

        serving_fn = self.create_serving_signature(
            model, feature_mapper, label_inverse_lookup_layer
        )

        saved_model_dir = tempfile.mkdtemp(dir=self.get_temp_dir())
        model.save(
            saved_model_dir,
            save_format="tf",
            signatures={"serving_default": serving_fn},
        )

        # Test the saved_model.
        loaded_serving_fn = keras.saving.save.load_model(
            saved_model_dir
        ).signatures["serving_default"]

        # check the result w/ and w/o avenger.
        prediction0 = loaded_serving_fn(
            tf.constant(["avenger", "ironman", "avenger"])
        )["output_0"]
        self.assertIn(prediction0.numpy().decode("UTF-8"), ("yes", "no"))

        prediction1 = loaded_serving_fn(
            tf.constant(["ironman", "ironman", "unknown"])
        )["output_0"]
        self.assertIn(prediction1.numpy().decode("UTF-8"), ("yes", "no"))



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Built-in optimizer classes.

For more examples see the base class `tf.keras.optimizers.Optimizer`.
"""

import tensorflow.compat.v2 as tf

# Imports needed for deserialization.
from keras import backend
from keras.optimizers.legacy import adadelta as adadelta_legacy
from keras.optimizers.legacy import adagrad as adagrad_legacy
from keras.optimizers.legacy import adam as adam_legacy
from keras.optimizers.legacy import adamax as adamax_legacy
from keras.optimizers.legacy import ftrl as ftrl_legacy
from keras.optimizers.legacy import nadam as nadam_legacy
from keras.optimizers.legacy import optimizer as optimizer_legacy
from keras.optimizers.legacy import rmsprop as rmsprop_legacy
from keras.optimizers.legacy import sgd as sgd_legacy
from keras.optimizers.optimizer_experimental import (
    adadelta as adadelta_experimental,
)
from keras.optimizers.optimizer_experimental import (
    adagrad as adagrad_experimental,
)
from keras.optimizers.optimizer_experimental import adam as adam_experimental
from keras.optimizers.optimizer_experimental import (
    adamax as adamax_experimental,
)
from keras.optimizers.optimizer_experimental import adamw as adamw_experimental
from keras.optimizers.optimizer_experimental import ftrl as ftrl_experimental
from keras.optimizers.optimizer_experimental import nadam as nadam_experimental
from keras.optimizers.optimizer_experimental import (
    optimizer as optimizer_experimental,
)
from keras.optimizers.optimizer_experimental import (
    rmsprop as rmsprop_experimental,
)
from keras.optimizers.optimizer_experimental import sgd as sgd_experimental
from keras.optimizers.optimizer_v1 import Optimizer
from keras.optimizers.optimizer_v1 import TFOptimizer
from keras.optimizers.optimizer_v2 import adadelta as adadelta_v2
from keras.optimizers.optimizer_v2 import adagrad as adagrad_v2
from keras.optimizers.optimizer_v2 import adam as adam_v2
from keras.optimizers.optimizer_v2 import adamax as adamax_v2
from keras.optimizers.optimizer_v2 import ftrl
from keras.optimizers.optimizer_v2 import (
    gradient_descent as gradient_descent_v2,
)
from keras.optimizers.optimizer_v2 import nadam as nadam_v2
from keras.optimizers.optimizer_v2 import optimizer_v2 as base_optimizer_v2
from keras.optimizers.optimizer_v2 import rmsprop as rmsprop_v2
from keras.optimizers.optimizer_v2.adadelta import Adadelta
from keras.optimizers.optimizer_v2.adagrad import Adagrad
from keras.optimizers.optimizer_v2.adam import Adam
from keras.optimizers.optimizer_v2.adamax import Adamax
from keras.optimizers.optimizer_v2.ftrl import Ftrl

# Symbols to be accessed under keras.optimizers. To be replaced with
# optimizers v2022 when they graduate out of experimental.
from keras.optimizers.optimizer_v2.gradient_descent import SGD
from keras.optimizers.optimizer_v2.nadam import Nadam
from keras.optimizers.optimizer_v2.rmsprop import RMSprop
from keras.utils.generic_utils import deserialize_keras_object
from keras.utils.generic_utils import serialize_keras_object

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.optimizers.serialize")
def serialize(optimizer):
    """Serialize the optimizer configuration to JSON compatible python dict.

    The configuration can be used for persistence and reconstruct the
    `Optimizer` instance again.

    >>> tf.keras.optimizers.serialize(tf.keras.optimizers.SGD())
    {'class_name': 'SGD', 'config': {'name': 'SGD', 'learning_rate': 0.01,
                                     'decay': 0.0, 'momentum': 0.0,
                                     'nesterov': False}}

    Args:
      optimizer: An `Optimizer` instance to serialize.

    Returns:
      Python dict which contains the configuration of the input optimizer.
    """
    return serialize_keras_object(optimizer)


@keras_export("keras.optimizers.deserialize")
def deserialize(config, custom_objects=None, **kwargs):
    """Inverse of the `serialize` function.

    Args:
        config: Optimizer configuration dictionary.
        custom_objects: Optional dictionary mapping names (strings) to custom
          objects (classes and functions) to be considered during
          deserialization.

    Returns:
        A Keras Optimizer instance.
    """
    # loss_scale_optimizer has a direct dependency of optimizer, import here
    # rather than top to avoid the cyclic dependency.
    from keras.mixed_precision import (
        loss_scale_optimizer,
    )

    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", True)
    if len(config["config"]) > 0:
        # If the optimizer config is not empty, then we use the value of
        # `is_legacy_optimizer` to override `use_legacy_optimizer`. If
        # `is_legacy_optimizer` does not exist in config, it means we are
        # using the legacy optimzier.
        use_legacy_optimizer = config["config"].get("is_legacy_optimizer", True)
    if (
        tf.__internal__.tf2.enabled()
        and tf.executing_eagerly()
        and not use_legacy_optimizer
    ):
        all_classes = {
            "adadelta": adadelta_experimental.Adadelta,
            "adagrad": adagrad_experimental.Adagrad,
            "adam": adam_experimental.Adam,
            "adamax": adamax_experimental.Adamax,
            "experimentaladadelta": adadelta_experimental.Adadelta,
            "experimentaladagrad": adagrad_experimental.Adagrad,
            "experimentaladam": adam_experimental.Adam,
            "experimentalsgd": sgd_experimental.SGD,
            "nadam": nadam_experimental.Nadam,
            "rmsprop": rmsprop_experimental.RMSprop,
            "sgd": sgd_experimental.SGD,
            "ftrl": ftrl_experimental.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizerV3,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }
    else:
        all_classes = {
            "adadelta": adadelta_v2.Adadelta,
            "adagrad": adagrad_v2.Adagrad,
            "adam": adam_v2.Adam,
            "adamax": adamax_v2.Adamax,
            "experimentaladadelta": adadelta_experimental.Adadelta,
            "experimentaladagrad": adagrad_experimental.Adagrad,
            "experimentaladam": adam_experimental.Adam,
            "experimentalsgd": sgd_experimental.SGD,
            "nadam": nadam_v2.Nadam,
            "rmsprop": rmsprop_v2.RMSprop,
            "sgd": gradient_descent_v2.SGD,
            "ftrl": ftrl.Ftrl,
            "lossscaleoptimizer": loss_scale_optimizer.LossScaleOptimizer,
            "lossscaleoptimizerv3": loss_scale_optimizer.LossScaleOptimizerV3,
            # LossScaleOptimizerV1 was an old version of LSO that was removed.
            # Deserializing it turns it into a LossScaleOptimizer
            "lossscaleoptimizerv1": loss_scale_optimizer.LossScaleOptimizer,
        }

    # Make deserialization case-insensitive for built-in optimizers.
    if config["class_name"].lower() in all_classes:
        config["class_name"] = config["class_name"].lower()
    return deserialize_keras_object(
        config,
        module_objects=all_classes,
        custom_objects=custom_objects,
        printable_module_name="optimizer",
    )


@keras_export("keras.optimizers.get")
def get(identifier, **kwargs):
    """Retrieves a Keras Optimizer instance.

    Args:
        identifier: Optimizer identifier, one of
            - String: name of an optimizer
            - Dictionary: configuration dictionary.
            - Keras Optimizer instance (it will be returned unchanged).
            - TensorFlow Optimizer instance (it will be wrapped as a Keras
              Optimizer).

    Returns:
        A Keras Optimizer instance.

    Raises:
        ValueError: If `identifier` cannot be interpreted.
    """
    use_legacy_optimizer = kwargs.pop("use_legacy_optimizer", True)
    if isinstance(
        identifier,
        (
            Optimizer,
            base_optimizer_v2.OptimizerV2,
            optimizer_experimental.Optimizer,
        ),
    ):
        return identifier
    # Wrap legacy TF optimizer instances
    elif isinstance(identifier, tf.compat.v1.train.Optimizer):
        opt = TFOptimizer(identifier)
        backend.track_tf_optimizer(opt)
        return opt
    elif isinstance(identifier, dict):
        return deserialize(
            identifier, use_legacy_optimizer=use_legacy_optimizer
        )
    elif isinstance(identifier, str):
        config = {"class_name": str(identifier), "config": {}}
        return deserialize(config, use_legacy_optimizer=use_legacy_optimizer)
    else:
        raise ValueError(
            f"Could not interpret optimizer identifier: {identifier}"
        )



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras upsampling layer for 3D inputs."""


import tensorflow.compat.v2 as tf

from keras import backend
from keras.engine.base_layer import Layer
from keras.engine.input_spec import InputSpec
from keras.utils import conv_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.UpSampling3D")
class UpSampling3D(Layer):
    """Upsampling layer for 3D inputs.

    Repeats the 1st, 2nd and 3rd dimensions
    of the data by `size[0]`, `size[1]` and `size[2]` respectively.

    Examples:

    >>> input_shape = (2, 1, 2, 1, 3)
    >>> x = tf.constant(1, shape=input_shape)
    >>> y = tf.keras.layers.UpSampling3D(size=2)(x)
    >>> print(y.shape)
    (2, 2, 4, 2, 3)

    Args:
      size: Int, or tuple of 3 integers.
        The upsampling factors for dim1, dim2 and dim3.
      data_format: A string,
        one of `channels_last` (default) or `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, spatial_dim1, spatial_dim2, spatial_dim3, channels)`
        while `channels_first` corresponds to inputs with shape
        `(batch_size, channels, spatial_dim1, spatial_dim2, spatial_dim3)`.
        It defaults to the `image_data_format` value found in your
        Keras config file at `~/.keras/keras.json`.
        If you never set it, then it will be "channels_last".

    Input shape:
      5D tensor with shape:
      - If `data_format` is `"channels_last"`:
          `(batch_size, dim1, dim2, dim3, channels)`
      - If `data_format` is `"channels_first"`:
          `(batch_size, channels, dim1, dim2, dim3)`

    Output shape:
      5D tensor with shape:
      - If `data_format` is `"channels_last"`:
          `(batch_size, upsampled_dim1, upsampled_dim2, upsampled_dim3,
          channels)`
      - If `data_format` is `"channels_first"`:
          `(batch_size, channels, upsampled_dim1, upsampled_dim2,
          upsampled_dim3)`
    """

    def __init__(self, size=(2, 2, 2), data_format=None, **kwargs):
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.size = conv_utils.normalize_tuple(size, 3, "size")
        self.input_spec = InputSpec(ndim=5)
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        if self.data_format == "channels_first":
            dim1 = (
                self.size[0] * input_shape[2]
                if input_shape[2] is not None
                else None
            )
            dim2 = (
                self.size[1] * input_shape[3]
                if input_shape[3] is not None
                else None
            )
            dim3 = (
                self.size[2] * input_shape[4]
                if input_shape[4] is not None
                else None
            )
            return tf.TensorShape(
                [input_shape[0], input_shape[1], dim1, dim2, dim3]
            )
        else:
            dim1 = (
                self.size[0] * input_shape[1]
                if input_shape[1] is not None
                else None
            )
            dim2 = (
                self.size[1] * input_shape[2]
                if input_shape[2] is not None
                else None
            )
            dim3 = (
                self.size[2] * input_shape[3]
                if input_shape[3] is not None
                else None
            )
            return tf.TensorShape(
                [input_shape[0], dim1, dim2, dim3, input_shape[4]]
            )

    def call(self, inputs):
        return backend.resize_volumes(
            inputs, self.size[0], self.size[1], self.size[2], self.data_format
        )

    def get_config(self):
        config = {"size": self.size, "data_format": self.data_format}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras Applications are premade architectures with pre-trained weights.
"""

import sys as _sys

from keras.api._v1.keras.applications import convnext
from keras.api._v1.keras.applications import densenet
from keras.api._v1.keras.applications import efficientnet
from keras.api._v1.keras.applications import efficientnet_v2
from keras.api._v1.keras.applications import imagenet_utils
from keras.api._v1.keras.applications import inception_resnet_v2
from keras.api._v1.keras.applications import inception_v3
from keras.api._v1.keras.applications import mobilenet
from keras.api._v1.keras.applications import mobilenet_v2
from keras.api._v1.keras.applications import mobilenet_v3
from keras.api._v1.keras.applications import nasnet
from keras.api._v1.keras.applications import regnet
from keras.api._v1.keras.applications import resnet
from keras.api._v1.keras.applications import resnet50
from keras.api._v1.keras.applications import resnet_rs
from keras.api._v1.keras.applications import resnet_v2
from keras.api._v1.keras.applications import vgg16
from keras.api._v1.keras.applications import vgg19
from keras.api._v1.keras.applications import xception
from keras.applications.convnext import ConvNeXtBase
from keras.applications.convnext import ConvNeXtLarge
from keras.applications.convnext import ConvNeXtSmall
from keras.applications.convnext import ConvNeXtTiny
from keras.applications.convnext import ConvNeXtXLarge
from keras.applications.densenet import DenseNet121
from keras.applications.densenet import DenseNet169
from keras.applications.densenet import DenseNet201
from keras.applications.efficientnet import EfficientNetB0
from keras.applications.efficientnet import EfficientNetB1
from keras.applications.efficientnet import EfficientNetB2
from keras.applications.efficientnet import EfficientNetB3
from keras.applications.efficientnet import EfficientNetB4
from keras.applications.efficientnet import EfficientNetB5
from keras.applications.efficientnet import EfficientNetB6
from keras.applications.efficientnet import EfficientNetB7
from keras.applications.efficientnet_v2 import EfficientNetV2B0
from keras.applications.efficientnet_v2 import EfficientNetV2B1
from keras.applications.efficientnet_v2 import EfficientNetV2B2
from keras.applications.efficientnet_v2 import EfficientNetV2B3
from keras.applications.efficientnet_v2 import EfficientNetV2L
from keras.applications.efficientnet_v2 import EfficientNetV2M
from keras.applications.efficientnet_v2 import EfficientNetV2S
from keras.applications.inception_resnet_v2 import InceptionResNetV2
from keras.applications.inception_v3 import InceptionV3
from keras.applications.mobilenet import MobileNet
from keras.applications.mobilenet_v2 import MobileNetV2
from keras.applications.mobilenet_v3 import MobileNetV3Large
from keras.applications.mobilenet_v3 import MobileNetV3Small
from keras.applications.nasnet import NASNetLarge
from keras.applications.nasnet import NASNetMobile
from keras.applications.regnet import RegNetX002
from keras.applications.regnet import RegNetX004
from keras.applications.regnet import RegNetX006
from keras.applications.regnet import RegNetX008
from keras.applications.regnet import RegNetX016
from keras.applications.regnet import RegNetX032
from keras.applications.regnet import RegNetX040
from keras.applications.regnet import RegNetX064
from keras.applications.regnet import RegNetX080
from keras.applications.regnet import RegNetX120
from keras.applications.regnet import RegNetX160
from keras.applications.regnet import RegNetX320
from keras.applications.regnet import RegNetY002
from keras.applications.regnet import RegNetY004
from keras.applications.regnet import RegNetY006
from keras.applications.regnet import RegNetY008
from keras.applications.regnet import RegNetY016
from keras.applications.regnet import RegNetY032
from keras.applications.regnet import RegNetY040
from keras.applications.regnet import RegNetY064
from keras.applications.regnet import RegNetY080
from keras.applications.regnet import RegNetY120
from keras.applications.regnet import RegNetY160
from keras.applications.regnet import RegNetY320
from keras.applications.resnet import ResNet101
from keras.applications.resnet import ResNet152
from keras.applications.resnet import ResNet50
from keras.applications.resnet_rs import ResNetRS101
from keras.applications.resnet_rs import ResNetRS152
from keras.applications.resnet_rs import ResNetRS200
from keras.applications.resnet_rs import ResNetRS270
from keras.applications.resnet_rs import ResNetRS350
from keras.applications.resnet_rs import ResNetRS420
from keras.applications.resnet_rs import ResNetRS50
from keras.applications.resnet_v2 import ResNet101V2
from keras.applications.resnet_v2 import ResNet152V2
from keras.applications.resnet_v2 import ResNet50V2
from keras.applications.vgg16 import VGG16
from keras.applications.vgg19 import VGG19
from keras.applications.xception import Xception
from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "keras.applications", public_apis=None, deprecation=True,
      has_lite=False)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for stateful tf.keras LSTM models using DistributionStrategy."""

import numpy as np
import tensorflow.compat.v2 as tf

import keras
from keras.distribute import keras_correctness_test_base
from keras.optimizers.optimizer_v2 import (
    gradient_descent as gradient_descent_keras,
)


def strategies_for_stateful_embedding_model():
    """Returns TPUStrategy with single core device assignment."""

    return [
        tf.__internal__.distribute.combinations.tpu_strategy_one_core,
    ]


def test_combinations_for_stateful_embedding_model():
    return tf.__internal__.test.combinations.combine(
        distribution=strategies_for_stateful_embedding_model(),
        mode="graph",
        use_numpy=False,
        use_validation_data=False,
    )


class DistributionStrategyStatefulLstmModelCorrectnessTest(
    keras_correctness_test_base.TestDistributionStrategyEmbeddingModelCorrectnessBase  # noqa: E501
):
    def get_model(
        self,
        max_words=10,
        initial_weights=None,
        distribution=None,
        input_shapes=None,
    ):
        del input_shapes
        batch_size = keras_correctness_test_base._GLOBAL_BATCH_SIZE

        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            word_ids = keras.layers.Input(
                shape=(max_words,),
                batch_size=batch_size,
                dtype=np.int32,
                name="words",
            )
            word_embed = keras.layers.Embedding(input_dim=20, output_dim=10)(
                word_ids
            )
            lstm_embed = keras.layers.LSTM(
                units=4, return_sequences=False, stateful=True
            )(word_embed)

            preds = keras.layers.Dense(2, activation="softmax")(lstm_embed)
            model = keras.Model(inputs=[word_ids], outputs=[preds])

            if initial_weights:
                model.set_weights(initial_weights)

            optimizer_fn = gradient_descent_keras.SGD

            model.compile(
                optimizer=optimizer_fn(learning_rate=0.1),
                loss="sparse_categorical_crossentropy",
                metrics=["sparse_categorical_accuracy"],
            )
        return model

    # TODO(jhseu): Disabled to fix b/130808953. Need to investigate why it
    # doesn't work and enable for DistributionStrategy more generally.
    @tf.__internal__.distribute.combinations.generate(
        test_combinations_for_stateful_embedding_model()
    )
    def disabled_test_stateful_lstm_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(
            distribution, use_numpy, use_validation_data, is_stateful_model=True
        )

    @tf.__internal__.distribute.combinations.generate(
        tf.__internal__.test.combinations.times(
            keras_correctness_test_base.test_combinations_with_tpu_strategies_graph()  # noqa: E501
        )
    )
    def test_incorrectly_use_multiple_cores_for_stateful_lstm_model(
        self, distribution, use_numpy, use_validation_data
    ):
        with self.assertRaisesRegex(
            ValueError, "not yet supported with tf.distribute.Strategy"
        ):
            self.run_correctness_test(
                distribution,
                use_numpy,
                use_validation_data,
                is_stateful_model=True,
            )


if __name__ == "__main__":
    tf.test.main()



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A `Network` is way to compose layers: the topological form of a `Model`."""


import collections
import copy
import itertools
import warnings

import tensorflow.compat.v2 as tf

from keras import backend
from keras.dtensor import layout_map as layout_map_lib
from keras.engine import base_layer
from keras.engine import base_layer_utils
from keras.engine import functional_utils
from keras.engine import input_layer as input_layer_module
from keras.engine import input_spec
from keras.engine import node as node_module
from keras.engine import training as training_lib
from keras.engine import training_utils
from keras.saving.saved_model import json_utils
from keras.saving.saved_model import network_serialization
from keras.utils import generic_utils
from keras.utils import tf_inspect
from keras.utils import tf_utils

# isort: off
from tensorflow.python.platform import tf_logging as logging
from tensorflow.tools.docs import doc_controls


class Functional(training_lib.Model):
    """A `Functional` model is a `Model` defined as a directed graph of layers.

    Three types of `Model` exist: subclassed `Model`, `Functional` model,
    and `Sequential` (a special case of `Functional`).
    In general, more Keras features are supported with `Functional`
    than with subclassed `Model`s, specifically:

    - Model cloning (`keras.models.clone`)
    - Serialization (`model.get_config()/from_config`, `model.to_json()`
    - Whole-model saving (`model.save()`)

    A `Functional` model can be instantiated by passing two arguments to
    `__init__`. The first argument is the `keras.Input` Tensors that represent
    the inputs to the model. The second argument specifies the output
    tensors that represent the outputs of this model. Both arguments can be a
    nested structure of tensors.

    Example:

    ```
    inputs = {'x1': keras.Input(shape=(10,)), 'x2': keras.Input(shape=(1,))}
    t = keras.layers.Dense(1, activation='relu')(inputs['x1'])
    outputs = keras.layers.Add()([t, inputs['x2'])
    model = keras.Model(inputs, outputs)
    ```

    A `Functional` model constructed using the Functional API can also include
    raw TensorFlow functions, with the exception of functions that create
    Variables or assign ops.

    Example:

    ```python
    inputs = keras.Input(shape=(10,))
    x = keras.layers.Dense(1)(inputs)
    outputs = tf.nn.relu(x)
    model = keras.Model(inputs, outputs)
    ```

    A new `Functional` model can also be created by using the
    intermediate tensors. This enables you to quickly extract sub-components
    of the model.

    Example:

    ```python
    inputs = keras.Input(shape=(None, None, 3))
    processed = keras.layers.RandomCrop(width=32, height=32)(inputs)
    conv = keras.layers.Conv2D(filters=2, kernel_size=3)(processed)
    pooling = keras.layers.GlobalAveragePooling2D()(conv)
    feature = keras.layers.Dense(10)(pooling)

    full_model = keras.Model(inputs, feature)
    backbone = keras.Model(processed, conv)
    activations = keras.Model(conv, feature)
    ```

    Note that the `backbone` and `activations` models are not
    created with `keras.Input` objects, but with the tensors that are originated
    from `keras.Input` objects. Under the hood, the layers and weights will
    be shared across these models, so that user can train the `full_model`, and
    use `backbone` or `activations` to do feature extraction.
    The inputs and outputs of the model can be nested structures of tensors as
    well, and the created models are standard `Functional` model that support
    all the existing API.

    Args:
      inputs: List of input tensors (must be created via `tf.keras.Input()` or
        originated from `tf.keras.Input()`).
      outputs: List of output tensors.
      name: String, optional. Name of the model.
      trainable: Boolean, optional. If the model's variables should be
        trainable.
    """

    # See tf.Module for the usage of this property.
    # The key of _layer_call_argspecs is a layer. tf.Module._flatten will fail
    # to flatten the key since it is trying to convert Trackable/Layer to a
    # string.
    _TF_MODULE_IGNORED_PROPERTIES = frozenset(
        itertools.chain(
            (
                "_layer_call_argspecs",
                "_compiled_trainable_state",
                "_output_mask_cache",
                "_output_tensor_cache",
                "_output_shape_cache",
            ),
            training_lib.Model._TF_MODULE_IGNORED_PROPERTIES,
        )
    )

    @tf.__internal__.tracking.no_automatic_dependency_tracking
    def __init__(self, inputs, outputs, name=None, trainable=True, **kwargs):
        # This is used by the Model class, since we have some logic to swap the
        # class in the __new__ method, which will lead to __init__ get invoked
        # twice. Using the skip_init to skip one of the invocation of __init__
        # to avoid any side effects
        skip_init = kwargs.pop("skip_init", False)
        if skip_init:
            return
        generic_utils.validate_kwargs(kwargs, {})
        super().__init__(name=name, trainable=trainable)
        # Check if the inputs contain any intermediate `KerasTensor` (not
        # created by tf.keras.Input()). In this case we need to clone the `Node`
        # and `KerasTensor` objects to mimic rebuilding a new model from new
        # inputs.  This feature is only enabled in TF2 not in v1 graph mode.
        if tf.compat.v1.executing_eagerly_outside_functions():
            if not all(
                [
                    functional_utils.is_input_keras_tensor(t)
                    for t in tf.nest.flatten(inputs)
                ]
            ):
                inputs, outputs = functional_utils.clone_graph_nodes(
                    inputs, outputs
                )
        self._init_graph_network(inputs, outputs)

    @tf.__internal__.tracking.no_automatic_dependency_tracking
    def _init_graph_network(self, inputs, outputs):
        # This method is needed for Sequential to reinitialize graph network
        # when layer is added or removed.

        base_layer.keras_api_gauge.get_cell("Functional").set(True)
        self._is_graph_network = True

        # Normalize and set self.inputs, self.outputs.
        if isinstance(inputs, list) and len(tf.nest.flatten(inputs)) == 1:
            inputs = inputs[0]
        if isinstance(outputs, list) and len(tf.nest.flatten(outputs)) == 1:
            outputs = outputs[0]
        self._nested_inputs = inputs
        self._nested_outputs = outputs
        self.inputs = tf.nest.flatten(inputs)
        self.outputs = tf.nest.flatten(outputs)

        # Models constructed with a single Tensor or list of Tensors can
        # be called with a dict, where the keys of the dict are the names
        # of the `Input` objects. Extra keys are ignored with warning.
        if not tf.nest.is_nested(self._nested_inputs):
            self._enable_dict_to_input_mapping = True
        elif isinstance(self._nested_inputs, (list, tuple)) and not any(
            tf.nest.is_nested(t) for t in self._nested_inputs
        ):
            self._enable_dict_to_input_mapping = True
        elif isinstance(self._nested_inputs, dict) and not any(
            tf.nest.is_nested(t) for t in self._nested_inputs.values()
        ):
            self._enable_dict_to_input_mapping = True
        else:
            self._enable_dict_to_input_mapping = False

        if not tf.compat.v1.executing_eagerly_outside_functions():
            if any(
                not hasattr(tensor, "_keras_history") for tensor in self.outputs
            ):
                base_layer_utils.create_keras_history(self._nested_outputs)

        self._validate_graph_inputs_and_outputs()

        # A Network does not create weights of its own, thus it is already
        # built.
        self.built = True
        self._build_input_shape = tf.nest.map_structure(
            lambda x: x.shape, inputs
        )
        self._compute_output_and_mask_jointly = True
        # `_expects_training_arg` is True since the `training` argument is
        # always present in the signature of the `call` method of a graph
        # network.
        self._call_spec.expects_training_arg = True
        self._call_spec.expects_mask_arg = True
        # A graph network does not autocast inputs, as its layers will cast them
        # instead.
        self._autocast = False

        self._input_layers = []
        self._output_layers = []
        self._input_coordinates = []
        self._output_coordinates = []

        # This is for performance optimization when calling the Network on new
        # inputs. Every time the Network is called on a set on input tensors, we
        # compute the output tensors, output masks and output shapes in one
        # pass, then cache them here. When any of these outputs is queried
        # later, we retrieve it from there instead of recomputing it.
        self._output_mask_cache = {}
        self._output_tensor_cache = {}
        self._output_shape_cache = {}

        # Build self._output_layers:
        for x in self.outputs:
            (
                layer,
                node_index,
                tensor_index,
            ) = x._keras_history
            self._output_layers.append(layer)
            self._output_coordinates.append((layer, node_index, tensor_index))

        # Build self._input_layers:
        for x in self.inputs:
            (
                layer,
                node_index,
                tensor_index,
            ) = x._keras_history
            # It's supposed to be an input layer, so only one node
            # and one tensor output.
            assert node_index == 0
            assert tensor_index == 0
            self._input_layers.append(layer)
            self._input_coordinates.append((layer, node_index, tensor_index))

        # Keep track of the network's nodes and layers.
        nodes, nodes_by_depth, layers, _ = _map_graph_network(
            self.inputs, self.outputs
        )
        self._network_nodes = nodes
        self._nodes_by_depth = nodes_by_depth
        self._self_tracked_trackables = layers
        self._layer_call_argspecs = {}
        for layer in self._self_tracked_trackables:
            self._layer_call_argspecs[layer] = tf_inspect.getfullargspec(
                layer.call
            )

        # Build self.input_names and self.output_names.
        self._set_output_names()
        self.input_names = []
        self._feed_input_names = []
        self._feed_inputs = []
        self._feed_input_shapes = []
        for layer in self._input_layers:
            self.input_names.append(layer.name)
            if layer.is_placeholder:
                self._feed_input_names.append(layer.name)
                # Use batch_input_shape here because non-eager composite tensors
                # may not have a shape attribute that's meaningful (sparse, for
                # instance, has a tensor that's non-constant and needs to be
                # fed). This means that input layers that create placeholders
                # will need to have the batch_input_shape attr to allow for
                # input shape validation.
                self._feed_input_shapes.append(layer._batch_input_shape)
                self._feed_inputs.append(layer.input)

        self._compute_tensor_usage_count()
        self._set_save_spec(self._nested_inputs)
        tf_utils.assert_no_legacy_layers(self.layers)

        # Note that this method is used by both functional and sequential
        # models, so we can't just have this method in functional.__init__,
        # which will miss the coverage of sequential model.
        if self._layout_map is not None:
            layout_map_lib._map_functional_model_variable(
                self, self._layout_map
            )

    @property
    def input(self):
        """Retrieves the input tensor(s) of a layer.

        Only applicable if the layer has exactly one input,
        i.e. if it is connected to one incoming layer.

        Returns:
            Input tensor or list of input tensors.

        Raises:
          RuntimeError: If called in Eager mode.
          AttributeError: If no inbound nodes are found.
        """
        return self._nested_inputs

    @property
    def input_shape(self):
        """Retrieves the input shape(s) of a layer.

        Only applicable if the layer has exactly one input,
        i.e. if it is connected to one incoming layer, or if all inputs
        have the same shape.

        Returns:
            Input shape, as an integer shape tuple
            (or list of shape tuples, one tuple per input tensor).

        Raises:
            AttributeError: if the layer has no defined input_shape.
            RuntimeError: if called in Eager mode.
        """
        return tf.nest.map_structure(backend.int_shape, self.input)

    @property
    def input_spec(self):
        if hasattr(self, "_manual_input_spec"):
            return self._manual_input_spec
        if isinstance(self._nested_inputs, (dict, list, tuple)) and len(
            self._nested_inputs
        ) != len(self.inputs):
            # Case where we have a nested structure.
            # In such a case we can't safely run any checks.
            return None
        if isinstance(self._nested_inputs, dict):
            # Case where `_nested_inputs` is a plain dict of Inputs.
            names = sorted(self._nested_inputs.keys())
            return [
                input_spec.InputSpec(
                    shape=shape_with_no_batch_size(self._nested_inputs[name]),
                    allow_last_axis_squeeze=True,
                    name=name,
                )
                for name in names
            ]
        else:
            # Single input, or list / tuple of inputs.
            # The data may be passed as a dict keyed by input name.
            return [
                input_spec.InputSpec(
                    shape=shape_with_no_batch_size(x),
                    allow_last_axis_squeeze=True,
                    name=x._keras_history.layer.name,
                )
                for x in self.inputs
            ]

    @input_spec.setter
    def input_spec(self, value):
        self._manual_input_spec = value

    @property
    def output(self):
        """Retrieves the output tensor(s) of a layer.

        Only applicable if the layer has exactly one output,
        i.e. if it is connected to one incoming layer.

        Returns:
          Output tensor or list of output tensors.

        Raises:
          AttributeError: if the layer is connected to more than one incoming
            layers.
          RuntimeError: if called in Eager mode.
        """
        return self._nested_outputs

    @property
    def output_shape(self):
        """Retrieves the output shape(s) of a layer.

        Only applicable if the layer has one output,
        or if all outputs have the same shape.

        Returns:
            Output shape, as an integer shape tuple
            (or list of shape tuples, one tuple per output tensor).

        Raises:
            AttributeError: if the layer has no defined output shape.
            RuntimeError: if called in Eager mode.
        """
        return tf.nest.map_structure(backend.int_shape, self.output)

    def _set_output_names(self):
        """Assigns unique names to the Network's outputs.

        Output layers with multiple output tensors would otherwise lead to
        duplicate names in self.output_names.
        """
        uniquified = []
        output_names = set()
        prefix_count = {}
        for layer in self._output_layers:
            proposal = layer.name
            while proposal in output_names:
                existing_count = prefix_count.get(layer.name, 1)
                proposal = "{}_{}".format(layer.name, existing_count)
                prefix_count[layer.name] = existing_count + 1
            output_names.add(proposal)
            uniquified.append(proposal)
        self.output_names = uniquified

    @property
    def _layer_checkpoint_dependencies(self):
        """Dictionary of layer dependencies to be included in the checkpoint."""
        weight_layer_index = 0

        dependencies = collections.OrderedDict()
        for layer_index, layer in enumerate(self.layers):
            try:
                if layer.weights:
                    # Keep a separate index for layers which have weights. This
                    # allows users to insert Layers without weights anywhere in
                    # the network without breaking checkpoints.
                    dependencies[
                        "layer_with_weights-%d" % weight_layer_index
                    ] = layer
                    weight_layer_index += 1
            except ValueError:
                # The layer might have weights, but may not be built yet. We
                # just treat it as layer without weight.
                pass

            # Even if it doesn't have weights, we should still track everything
            # in case it has/will have Trackable dependencies.
            dependencies["layer-%d" % layer_index] = layer
        return dependencies

    def _trackable_children(self, save_type="checkpoint", **kwargs):
        dependencies = self._layer_checkpoint_dependencies
        dependencies.update(super()._trackable_children(save_type, **kwargs))
        return dependencies

    def _lookup_dependency(self, name):
        layer_dependencies = self._layer_checkpoint_dependencies
        if name in layer_dependencies:
            return layer_dependencies[name]
        return super()._lookup_dependency(name)

    def _handle_deferred_layer_dependencies(self, layers):
        """Handles layer checkpoint dependencies that are added after init."""
        layer_checkpoint_dependencies = self._layer_checkpoint_dependencies
        layer_to_name = {v: k for k, v in layer_checkpoint_dependencies.items()}
        for layer in layers:
            if layer in layer_to_name:
                self._handle_deferred_dependencies(
                    name=layer_to_name[layer], trackable=layer
                )

    @property
    def _should_compute_mask(self):
        return True

    def compute_mask(self, inputs, mask):
        # TODO(omalleyt): b/123540974 This function is not really safe to call
        # by itself because it will duplicate any updates and losses in graph
        # mode by `call`ing the Layers again.
        output_tensors = self._run_internal_graph(inputs, mask=mask)
        return tf.nest.map_structure(
            lambda t: getattr(t, "_keras_mask", None), output_tensors
        )

    @doc_controls.do_not_doc_inheritable
    def call(self, inputs, training=None, mask=None):
        """Calls the model on new inputs.

        In this case `call` just reapplies
        all ops in the graph to the new inputs
        (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: A tensor or list of tensors.
            training: Boolean or boolean scalar tensor, indicating whether to
                run the `Network` in training mode or inference mode.
            mask: A mask or list of masks. A mask can be
                either a tensor or None (no mask).

        Returns:
            A tensor if there is a single output, or
            a list of tensors if there are more than one outputs.
        """
        return self._run_internal_graph(inputs, training=training, mask=mask)

    def compute_output_shape(self, input_shape):
        # Convert any shapes in tuple format to TensorShapes.
        input_shape = tf_utils.convert_shapes(input_shape, to_tuples=False)

        if len(tf.nest.flatten(input_shape)) != len(
            tf.nest.flatten(self._input_layers)
        ):
            raise ValueError(
                f"Invalid `input_shape` argument {input_shape}: "
                f"the model expects {len(self._input_layers)} "
                "input tensors."
            )

        # Use the tuple of TensorShape as the cache key, since tuple is hashable
        # and can be used as hash key.
        try:
            cache_key = tuple(
                tf_utils.convert_shapes(input_shape, to_tuples=True)
            )
            if cache_key in self._output_shape_cache:
                # Cache hit. Return shapes as TensorShapes.
                return self._output_shape_cache[cache_key]
        except ValueError:
            # In case there are unknown TensorShape, eg for sparse tensor input,
            # We skip the caching since the shape is unknown.
            pass

        layers_to_output_shapes = {}
        for layer, shape in zip(
            self._input_layers, tf.nest.flatten(input_shape)
        ):
            # It's an input layer: then `compute_output_shape` is identity,
            # and there is only one node and one tensor..
            shape_key = layer.name + "_0_0"
            layers_to_output_shapes[shape_key] = shape

        depth_keys = list(self._nodes_by_depth.keys())
        depth_keys.sort(reverse=True)
        # Iterate over nodes, by depth level.
        if len(depth_keys) > 1:
            for depth in depth_keys:
                nodes = self._nodes_by_depth[depth]
                for node in nodes:
                    layer = node.layer
                    if layer in self._input_layers:
                        # We've already covered the input layers
                        # a few lines above.
                        continue
                    # Get the input shapes for the first argument of the node
                    layer_input_shapes = []
                    layer_inputs = node.call_args[0]
                    for layer_input in tf.nest.flatten(layer_inputs):
                        kh = layer_input._keras_history
                        input_layer_key = kh.layer.name + "_%s_%s" % (
                            kh.node_index,
                            kh.tensor_index,
                        )
                        layer_input_shapes.append(
                            layers_to_output_shapes[input_layer_key]
                        )
                    layer_input_shapes = tf.nest.pack_sequence_as(
                        layer_inputs, layer_input_shapes
                    )
                    # Layers expect shapes to be tuples for
                    # `compute_output_shape`.
                    layer_input_shapes = tf_utils.convert_shapes(
                        layer_input_shapes, to_tuples=True
                    )
                    layer_output_shapes = layer.compute_output_shape(
                        layer_input_shapes
                    )
                    # Convert back to TensorShapes.
                    layer_output_shapes = tf_utils.convert_shapes(
                        layer_output_shapes, to_tuples=False
                    )

                    node_index = layer._inbound_nodes.index(node)
                    for j, shape in enumerate(
                        tf.nest.flatten(layer_output_shapes)
                    ):
                        shape_key = layer.name + "_%s_%s" % (node_index, j)
                        layers_to_output_shapes[shape_key] = shape

            # Read final output shapes from layers_to_output_shapes.
            output_shapes = []
            for i in range(len(self._output_layers)):
                layer, node_index, tensor_index = self._output_coordinates[i]
                shape_key = layer.name + "_%s_%s" % (node_index, tensor_index)
                output_shapes.append(layers_to_output_shapes[shape_key])
            output_shapes = tf.nest.pack_sequence_as(
                self._nested_outputs, output_shapes
            )
            # Store in cache.
            self._output_shape_cache[cache_key] = output_shapes

        # Return shapes as TensorShapes.
        return output_shapes

    def _init_set_name(self, name, zero_based=True):
        if not name:
            cls_name = self.__class__.__name__
            if self.__class__ == Functional:
                # Hide the functional class name from user, since its not a
                # public visible class. Use "Model" instead,
                cls_name = "Model"
            self._name = backend.unique_object_name(
                generic_utils.to_snake_case(cls_name), zero_based=zero_based
            )
        else:
            self._name = name

    def _run_internal_graph(self, inputs, training=None, mask=None):
        """Computes output tensors for new inputs.

        # Note:
            - Can be run on non-Keras tensors.

        Args:
            inputs: Tensor or nested structure of Tensors.
            training: Boolean learning phase.
            mask: (Optional) Tensor or nested structure of Tensors.

        Returns:
            output_tensors
        """
        inputs = self._flatten_to_reference_inputs(inputs)
        if mask is None:
            masks = [None] * len(inputs)
        else:
            masks = self._flatten_to_reference_inputs(mask)
        for input_t, mask in zip(inputs, masks):
            input_t._keras_mask = mask

        # Dictionary mapping reference tensors to computed tensors.
        tensor_dict = {}
        tensor_usage_count = self._tensor_usage_count
        for x, y in zip(self.inputs, inputs):
            y = self._conform_to_reference_input(y, ref_input=x)
            x_id = str(id(x))
            tensor_dict[x_id] = [y] * tensor_usage_count[x_id]

        nodes_by_depth = self._nodes_by_depth
        depth_keys = list(nodes_by_depth.keys())
        depth_keys.sort(reverse=True)

        for depth in depth_keys:
            nodes = nodes_by_depth[depth]
            for node in nodes:
                if node.is_input:
                    continue  # Input tensors already exist.

                if any(t_id not in tensor_dict for t_id in node.flat_input_ids):
                    continue  # Node is not computable, try skipping.

                args, kwargs = node.map_arguments(tensor_dict)
                outputs = node.layer(*args, **kwargs)

                # Update tensor_dict.
                for x_id, y in zip(
                    node.flat_output_ids, tf.nest.flatten(outputs)
                ):
                    tensor_dict[x_id] = [y] * tensor_usage_count[x_id]

        output_tensors = []
        for x in self.outputs:
            x_id = str(id(x))
            assert x_id in tensor_dict, "Could not compute output " + str(x)
            output_tensors.append(tensor_dict[x_id].pop())

        return tf.nest.pack_sequence_as(self._nested_outputs, output_tensors)

    def _flatten_to_reference_inputs(self, tensors):
        """Maps `tensors` to their respective `keras.Input`."""
        if self._enable_dict_to_input_mapping and isinstance(tensors, dict):
            ref_inputs = self._nested_inputs
            if not tf.nest.is_nested(ref_inputs):
                ref_inputs = [self._nested_inputs]
            if isinstance(ref_inputs, dict):
                # In the case that the graph is constructed with dict input
                # tensors, We will use the original dict key to map with the
                # keys in the input data. Note that the model.inputs is using
                # nest.flatten to process the input tensors, which means the
                # dict input tensors are ordered by their keys.
                ref_input_names = sorted(ref_inputs.keys())
            else:
                ref_input_names = [
                    inp._keras_history.layer.name for inp in ref_inputs
                ]

            # Raise an warning if there are more input data comparing to input
            # tensor
            if len(tensors) > len(ref_input_names):
                warnings.warn(
                    "Input dict contained keys {} which did not match any "
                    "model input. They will be ignored by the model.".format(
                        [n for n in tensors.keys() if n not in ref_input_names]
                    ),
                    stacklevel=2,
                )

            try:
                # Flatten in the order `Input`s were passed during Model
                # construction.
                return [tensors[n] for n in ref_input_names]
            except KeyError:
                # TODO(b/151582614)
                return tf.nest.flatten(tensors)

        # Otherwise both self.inputs and tensors will already be in same order.
        return tf.nest.flatten(tensors)

    def _conform_to_reference_input(self, tensor, ref_input):
        """Set shape and dtype based on `keras.Input`s."""
        if isinstance(tensor, tf.Tensor):
            # Allow (None,) and (None, 1) Tensors to be passed interchangeably.
            # Use the shape specified by the `keras.Input`.
            t_shape = tensor.shape
            t_rank = t_shape.rank
            ref_shape = ref_input.shape
            ref_rank = ref_shape.rank
            keras_history = getattr(tensor, "_keras_history", None)
            if t_rank is not None and ref_rank is not None:
                # Should squeeze last dimension.  True if tensor is (BATCH, ...,
                # 1) and reference is (BATCH, ...).
                if t_rank == ref_rank + 1 and t_shape[-1] == 1:
                    tensor = tf.squeeze(tensor, axis=-1)
                # Should expand last_dimension.  True if tensor is (BATCH, ...)
                # and reference is (BATCH, ..., 1).
                elif t_rank == ref_rank - 1 and ref_shape[-1] == 1:
                    tensor = tf.expand_dims(tensor, axis=-1)
            if keras_history is not None:  # Restore keras history.
                tensor._keras_history = keras_history

            # Add shape hints to Tensors that may have None shape dims but have
            # shapes defined by the `keras.Input` (not applicable in eager
            # mode).
            if not tf.executing_eagerly():
                try:
                    tensor.set_shape(tensor.shape.merge_with(ref_input.shape))
                except ValueError:
                    logging.warning(
                        "Model was constructed with shape {} for input {}, "
                        "but it was called on an input with incompatible "
                        "shape {}.".format(
                            ref_input.shape, ref_input, tensor.shape
                        )
                    )

            # Dtype casting.
            tensor = tf.cast(tensor, dtype=ref_input.dtype)
        elif tf_utils.is_extension_type(tensor):
            # Dtype casting (If the extension type has a non-variant dtype and
            # supports being cast).  Only cast if necessary (since some
            # extension types may not implement tf.cast).
            tensor_dtype = getattr(tensor, "dtype", None)
            ref_input_dtype = getattr(ref_input, "dtype", None)
            if (
                ref_input_dtype is not None
                and tensor_dtype is not None
                and tensor_dtype != ref_input_dtype
                and ref_input_dtype != tf.variant
            ):
                tensor = tf.cast(tensor, dtype=ref_input_dtype)

        return tensor

    def get_config(self):
        # Continue adding configs into what the super class has added.
        config = super().get_config()
        return copy.deepcopy(get_network_config(self, config=config))

    def get_weight_paths(self):
        result = {}
        for layer in self.layers:
            (
                descendants,
                object_paths_dict,
            ) = tf.__internal__.tracking.ObjectGraphView(
                layer
            ).breadth_first_traversal()
            for descendant in descendants:
                if isinstance(descendant, tf.Variable):
                    trackable_references = object_paths_dict[descendant]
                    object_path = ".".join(
                        [t.name for t in trackable_references]
                    )
                    result[layer.name + "." + object_path] = descendant
        return result

    def _validate_graph_inputs_and_outputs(self):
        """Validates the inputs and outputs of a Graph Network."""
        # Check for redundancy in inputs.
        if len({id(i) for i in self.inputs}) != len(self.inputs):
            raise ValueError(
                "The list of inputs passed to the model "
                "contains the same input multiple times. "
                "All inputs should only appear once."
                f"Received inputs={self.inputs}"
            )

        for x in self.inputs:
            # Check that x has appropriate `_keras_history` metadata.
            if not hasattr(x, "_keras_history"):
                cls_name = self.__class__.__name__
                raise ValueError(
                    f"Input tensors to a {cls_name} model "
                    "must come from `tf.keras.Input`. "
                    f"Received inputs={x} (missing previous layer metadata)."
                )
            # Check that x is an input tensor.

            layer = x._keras_history.layer
            if len(layer._inbound_nodes) > 1 or (
                layer._inbound_nodes and not layer._inbound_nodes[0].is_input
            ):
                cls_name = self.__class__.__name__
                logging.warning(
                    f"{cls_name} model inputs must come from "
                    "`tf.keras.Input` (thus holding past layer metadata). "
                    "They cannot be the output of "
                    "a previous non-Input layer. "
                    "Here, a tensor specified as "
                    f'input to "{self.name}" was not an Input tensor, '
                    f'it was generated by layer "{layer.name}".\n'
                    "Note that input tensors are "
                    "instantiated via `tensor = tf.keras.Input(shape)`.\n"
                    f"The tensor that caused the issue was: {x}"
                )

        # Check compatibility of batch sizes of Input Layers.
        input_batch_sizes = set(
            [
                training_utils.get_static_batch_size(x._keras_history.layer)
                for x in self.inputs
            ]
        )
        input_batch_sizes.discard(None)
        if len(input_batch_sizes) > 1:
            logging.warning(
                "Found incompatible static batch sizes among the "
                f"inputs. Batch sizes: {sorted(input_batch_sizes)}"
            )

        for x in self.outputs:
            if not hasattr(x, "_keras_history"):
                cls_name = self.__class__.__name__
                raise ValueError(
                    f"Output tensors of a {cls_name} model must be "
                    "the output of a TensorFlow `Layer` "
                    f"(thus holding past layer metadata). Found: {x}"
                )

    def _insert_layers(self, layers, relevant_nodes=None):
        """Inserts Layers into the Network after Network creation.

        This is only valid for Keras Graph Networks.  Layers added via this
        function will be included in the `call` computation and `get_config` of
        this Network.  They will not be added to the Network's outputs.

        Args:
          layers: Arbitrary nested structure of Layers. Layers must be reachable
            from one or more of the `keras.Input` Tensors that correspond to
            this Network's inputs.
          relevant_nodes: Nodes from the Layers that should be considered part
            of this Network. If `None`, all Nodes will be considered part of
            this Network.

        Raises:
          ValueError: If the layers depend on `Input`s not found in this Model.
        """
        layers = tf.nest.flatten(layers)
        tf_utils.assert_no_legacy_layers(layers)
        node_to_depth = {}
        for depth, nodes in self._nodes_by_depth.items():
            node_to_depth.update({node: depth for node in nodes})
        # The nodes of these Layers that are relevant to this Network. If not
        # provided, assume all Nodes are relevant
        if not relevant_nodes:
            relevant_nodes = tf.nest.flatten(
                [layer._inbound_nodes for layer in layers]
            )
        network_nodes = set(relevant_nodes + list(node_to_depth.keys()))

        def _get_min_depth(node):
            """Gets the minimum depth at which node can be computed."""
            min_depth = 0
            for layer, node_id, _, _ in node.iterate_inbound():
                inbound_node = layer._inbound_nodes[node_id]
                if inbound_node in node_to_depth:
                    min_depth = min(min_depth, node_to_depth[inbound_node])
                elif inbound_node not in network_nodes:
                    continue
                else:
                    # Previous relevant nodes haven't been processed yet.
                    return None
            # New node is one shallower than its shallowest input.
            return min_depth - 1

        # Insert nodes into `_nodes_by_depth` and other node attrs.
        unprocessed_nodes = copy.copy(relevant_nodes)
        i = 0
        while unprocessed_nodes:
            i += 1
            # Do a sanity check. This can occur if `Input`s from outside this
            # Model are being relied on.
            if i > 10000:
                raise ValueError(
                    "Layers could not be added due to missing " "dependencies."
                )

            node = unprocessed_nodes.pop(0)
            depth = _get_min_depth(node)
            if depth is None:  # Defer until inbound nodes are processed.
                unprocessed_nodes.append(node)
                continue
            node_key = _make_node_key(
                node.layer.name, node.layer._inbound_nodes.index(node)
            )
            if node_key not in self._network_nodes:
                node_to_depth[node] = depth
                self._network_nodes.add(node_key)
                self._nodes_by_depth[depth].append(node)

        # Insert layers and update other layer attrs.
        layer_set = set(self._self_tracked_trackables)
        deferred_layers = []
        for layer in layers:
            if layer not in layer_set:
                self._self_tracked_trackables.append(layer)
                deferred_layers.append(layer)
                self._layer_call_argspecs[layer] = tf_inspect.getfullargspec(
                    layer.call
                )
                layer_set.add(layer)
        self._handle_deferred_layer_dependencies(deferred_layers)

        self._compute_tensor_usage_count()

    def _compute_tensor_usage_count(self):
        """Compute the #. of tensor usages for all the output tensors of layers.

        The computed tensor usage count is saved as `self._tensor_usage_count`.
        This is later used for saving memory in eager computation by releasing
        no-longer-needed tensors as early as possible.
        """
        tensor_usage_count = collections.Counter()
        available_tensors = set(str(id(tensor)) for tensor in self.inputs)

        depth_keys = list(self._nodes_by_depth.keys())
        depth_keys.sort(reverse=True)
        depth_keys = depth_keys[1:]

        for depth in depth_keys:
            for node in self._nodes_by_depth[depth]:
                input_tensors = {
                    str(id(tensor))
                    for tensor in tf.nest.flatten(node.keras_inputs)
                }
                if input_tensors.issubset(available_tensors):
                    for tensor in tf.nest.flatten(node.keras_inputs):
                        tensor_usage_count[str(id(tensor))] += 1

                    for output_tensor in tf.nest.flatten(node.outputs):
                        available_tensors.add(str(id(output_tensor)))

        for tensor in self.outputs:
            tensor_usage_count[str(id(tensor))] += 1

        self._tensor_usage_count = tensor_usage_count

    def _assert_weights_created(self):
        # Override the implementation in Model.
        # The Functional model should always have weight created already.
        return

    def _graph_network_add_loss(self, symbolic_loss):
        new_nodes, new_layers = _map_subgraph_network(
            self.inputs, [symbolic_loss]
        )
        # Losses must be keyed on inputs no matter what in order to be supported
        # in DistributionStrategy.
        add_loss_layer = base_layer.AddLoss(
            unconditional=False, dtype=symbolic_loss.dtype
        )
        add_loss_layer(symbolic_loss)
        new_nodes.extend(add_loss_layer.inbound_nodes)
        new_layers.append(add_loss_layer)
        self._insert_layers(new_layers, new_nodes)

    def _graph_network_add_metric(self, value, aggregation, name):
        new_nodes, new_layers = _map_subgraph_network(self.inputs, [value])
        add_metric_layer = base_layer.AddMetric(
            aggregation, name, dtype=value.dtype
        )
        add_metric_layer(value)
        new_nodes.extend(add_metric_layer.inbound_nodes)
        new_layers.append(add_metric_layer)
        self._insert_layers(new_layers, new_nodes)

    @property
    def _trackable_saved_model_saver(self):
        return network_serialization.NetworkSavedModelSaver(self)

    def _get_save_spec(self, dynamic_batch=True, inputs_only=True):
        if getattr(self, "_has_explicit_input_shape", True):
            # Functional models and Sequential models that have an explicit
            # input shape should use the batch size set by the input layer.
            dynamic_batch = False
        return super()._get_save_spec(dynamic_batch, inputs_only)


def _make_node_key(layer_name, node_index):
    return layer_name + "_ib-" + str(node_index)


def _map_graph_network(inputs, outputs):
    """Validates a network's topology and gather its layers and nodes.

    Args:
      inputs: List of input tensors.
      outputs: List of outputs tensors.

    Returns:
      A tuple `(nodes, nodes_by_depth, layers, layers_by_depth)`.
      - nodes: list of Node instances.
      - nodes_by_depth: dict mapping ints (depth) to lists of node instances.
      - layers: list of Layer instances.
      - layers_by_depth: dict mapping ints (depth) to lists of layer instances.

    Raises:
      ValueError: In case the network is not valid (e.g. disconnected graph).
    """
    # "depth" is number of layers between output Node and the Node.
    # Nodes are ordered from inputs -> outputs.
    nodes_in_decreasing_depth, layer_indices = _build_map(outputs)
    network_nodes = {
        _make_node_key(node.layer.name, node.layer._inbound_nodes.index(node))
        for node in nodes_in_decreasing_depth
    }

    nodes_depths = {}  # dict {node: depth value}
    layers_depths = {}  # dict {layer: depth value}

    for node in reversed(nodes_in_decreasing_depth):
        # If the depth is not set, the node has no outbound nodes (depth 0).
        depth = nodes_depths.setdefault(node, 0)

        # Update the depth of the corresponding layer
        previous_depth = layers_depths.get(node.layer, 0)
        # If we've seen this layer before at a higher depth,
        # we should use that depth instead of the node depth.
        # This is necessary for shared layers that have inputs at different
        # depth levels in the graph.
        depth = max(depth, previous_depth)
        layers_depths[node.layer] = depth
        nodes_depths[node] = depth

        # Update the depth of inbound nodes.
        # The "depth" of a node is the max of the depths
        # of all nodes it is connected to + 1.
        for node_dep in node.parent_nodes:
            previous_depth = nodes_depths.get(node_dep, 0)
            nodes_depths[node_dep] = max(depth + 1, previous_depth)

    # Handle inputs that are not connected to outputs.
    # We do not error out here because the inputs may be used to compute losses
    # and metrics.
    for input_t in inputs:
        input_layer = input_t._keras_history[0]
        if input_layer not in layers_depths:
            layers_depths[input_layer] = 0
            layer_indices[input_layer] = -1
            nodes_depths[input_layer._inbound_nodes[0]] = 0
            network_nodes.add(_make_node_key(input_layer.name, 0))

    # Build a dict {depth: list of nodes with this depth}
    nodes_by_depth = collections.defaultdict(list)
    for node, depth in nodes_depths.items():
        nodes_by_depth[depth].append(node)

    # Build a dict {depth: list of layers with this depth}
    layers_by_depth = collections.defaultdict(list)
    for layer, depth in layers_depths.items():
        layers_by_depth[depth].append(layer)

    # Get sorted list of layer depths.
    depth_keys = list(layers_by_depth.keys())
    depth_keys.sort(reverse=True)

    # Set self.layers ordered by depth.
    layers = []
    for depth in depth_keys:
        layers_for_depth = layers_by_depth[depth]
        # Network.layers needs to have a deterministic order:
        # here we order them by traversal order.
        layers_for_depth.sort(key=lambda x: layer_indices[x])
        layers.extend(layers_for_depth)

    # Get sorted list of node depths.
    depth_keys = list(nodes_by_depth.keys())
    depth_keys.sort(reverse=True)

    # Check that all tensors required are computable.
    # computable_tensors: all tensors in the graph
    # that can be computed from the inputs provided.
    computable_tensors = set()
    for x in inputs:
        computable_tensors.add(id(x))

    layers_with_complete_input = []  # To provide a better error msg.
    for depth in depth_keys:
        for node in nodes_by_depth[depth]:
            layer = node.layer
            if layer and not node.is_input:
                for x in tf.nest.flatten(node.keras_inputs):
                    if id(x) not in computable_tensors:
                        raise ValueError(
                            f"Graph disconnected: cannot obtain value for "
                            f'tensor {x} at layer "{layer.name}". '
                            "The following previous layers were accessed "
                            f"without issue: {layers_with_complete_input}"
                        )
                for x in tf.nest.flatten(node.outputs):
                    computable_tensors.add(id(x))
                layers_with_complete_input.append(layer.name)

    # Ensure name unicity, which will be crucial for serialization
    # (since serialized nodes refer to layers by their name).
    all_names = [layer.name for layer in layers]
    for name in all_names:
        if all_names.count(name) != 1:
            raise ValueError(
                f'The name "{name}" is used {all_names.count(name)} '
                "times in the model. All layer names should be unique."
            )
    return network_nodes, nodes_by_depth, layers, layers_by_depth


def _build_map(outputs):
    """This method topologically sorts nodes in order from inputs to outputs.

    It uses a depth-first search to topologically sort nodes that appear in the
    _keras_history connectivity metadata of `outputs`.

    Args:
      outputs: the output tensors whose _keras_history metadata should be
        walked. This may be an arbitrary nested structure.

    Returns:
      A tuple like (ordered_nodes, layer_to_first_traversal_index)
      ordered_nodes: list of nodes appearing in the keras history, topologically
        sorted from original inputs to the `outputs`.
        (If outputs have different sets of ancestors, the inputs to one output
        may appear after a different output).
      layer_to_first_traversal_index:
        A dict mapping layer to the traversal index in the DFS where it is
        seen. Note: if a layer is shared by several nodes, the dict will only
        store the index corresponding to the *first* time the layer seen.
    """
    finished_nodes = set()
    nodes_in_progress = set()
    nodes_in_decreasing_depth = []  # nodes from inputs -> outputs.
    layer_indices = {}  # layer -> in traversal order.
    for output in tf.nest.flatten(outputs):
        _build_map_helper(
            output,
            finished_nodes,
            nodes_in_progress,
            nodes_in_decreasing_depth,
            layer_indices,
        )
    return nodes_in_decreasing_depth, layer_indices


def _build_map_helper(
    tensor,
    finished_nodes,
    nodes_in_progress,
    nodes_in_decreasing_depth,
    layer_indices,
):
    """Recursive helper for `_build_map`."""
    (
        layer,
        node_index,
        _,
    ) = tensor._keras_history
    node = layer._inbound_nodes[node_index]

    # Don't repeat work for shared subgraphs
    if node in finished_nodes:
        return

    # Prevent cycles.
    if node in nodes_in_progress:
        raise ValueError(
            f'Tensor {tensor} from layer "{layer.name}" ' "is part of a cycle."
        )

    # Store the traversal order for layer sorting.
    if layer not in layer_indices:
        layer_indices[layer] = len(layer_indices)

    # Propagate to all previous tensors connected to this node.
    nodes_in_progress.add(node)
    if not node.is_input:
        for tensor in node.keras_inputs:
            _build_map_helper(
                tensor,
                finished_nodes,
                nodes_in_progress,
                nodes_in_decreasing_depth,
                layer_indices,
            )

    finished_nodes.add(node)
    nodes_in_progress.remove(node)
    nodes_in_decreasing_depth.append(node)


def _map_subgraph_network(inputs, outputs):
    """Returns the nodes and layers in the topology from `inputs` to `outputs`.

    Args:
      inputs: List of input tensors.
      outputs: List of output tensors.

    Returns:
      A tuple of List{Node] and List[Layer].
    """
    if not tf.compat.v1.executing_eagerly_outside_functions():
        base_layer_utils.create_keras_history(outputs)
    # Keep only nodes and layers in the topology between inputs and outputs.
    _, nodes_by_depth, layers, _ = _map_graph_network(inputs, outputs)
    return tf.nest.flatten([nodes for nodes in nodes_by_depth.values()]), layers


def _should_skip_first_node(layer):
    """Returns True if the first layer node should not be saved or loaded."""
    # Networks that are constructed with an Input layer/shape start with a
    # pre-existing node linking their input to output. This node is excluded
    # from the network config.
    if layer._self_tracked_trackables:
        return (
            isinstance(layer, Functional)
            # Filter out Sequential models without an input shape.
            and isinstance(
                layer._self_tracked_trackables[0], input_layer_module.InputLayer
            )
        )
    else:
        return isinstance(layer, Functional)


def connect_ancillary_layers(model, created_layers):
    """Adds layers that are not connected to the outputs to the model."""
    # Layers not connected to outputs, such as those added in `add_loss`.
    ancillary_layers = [
        layer for layer in created_layers.values() if layer not in model.layers
    ]
    if ancillary_layers:
        relevant_nodes = tf.nest.flatten(
            [
                layer.inbound_nodes[1:]
                if _should_skip_first_node(layer)
                else layer.inbound_nodes
                for layer in created_layers.values()
            ]
        )
        model._insert_layers(ancillary_layers, relevant_nodes)
    return model


def reconstruct_from_config(config, custom_objects=None, created_layers=None):
    """Reconstructs graph from config object.

    Args:
      config: Dictionary returned from Network.get_config()
      custom_objects: Optional dictionary mapping names (strings) to custom
        classes or functions to be considered during deserialization.
      created_layers: Optional dictionary mapping names to Layer objects. Any
        layer not in this dictionary will be created and added to the dict.
        This function will add new nodes to all layers (excluding InputLayers),
        instead of re-using pre-existing nodes in the layers.

    Returns:
      Tuple of (input tensors, output tensors, dictionary of created layers)
    """
    # Layer instances created during the graph reconstruction process.
    created_layers = created_layers or collections.OrderedDict()

    # Maps input data (tuple of inbound layer name, node index) from the config
    # to node indices in the newly generated model. The node indices may be
    # different if the layers have already been called previously.
    node_index_map = {}
    node_count_by_layer = {}

    # Dictionary mapping layer instances to
    # node data that specifies a layer call.
    # It acts as a queue that maintains any unprocessed
    # layer call until it becomes possible to process it
    # (i.e. until the input tensors to the call all exist).
    unprocessed_nodes = collections.defaultdict(list)

    def get_node_index(layer, config_node_index):
        """Returns node index in layer (might differ from config_node_index)."""
        if isinstance(layer, input_layer_module.InputLayer):
            return 0
        return node_index_map.get((layer.name, config_node_index), None)

    def _deserialize_keras_tensors(kwargs, layer_map):
        """Deserializes Keras Tensors passed to `call`.."""

        def _deserialize_keras_tensor(t):
            """Deserializes a single Keras Tensor passed to `call`."""
            if isinstance(t, tf_utils.ListWrapper):
                t = t.as_list()
                layer_name = t[0]
                node_index = t[1]
                tensor_index = t[2]

                layer = layer_map[layer_name]
                new_node_index = get_node_index(layer, node_index)
                if new_node_index is None:
                    # The inbound node may not have been processed yet,
                    # (This can happen e.g. if it depends on a different set
                    # of inputs than those that have been processed already).
                    # raise an IndexError so that the current node puts itself
                    # back on the unprocessed queue.
                    # Caution: This may lead to infinite loops for malformed
                    # network configurations! (or when there is a bug in
                    # the network config loading code).
                    raise IndexError
                node = layer._inbound_nodes[new_node_index]
                return tf.nest.flatten(node.outputs)[tensor_index]
            return t

        kwargs = tf_utils.convert_inner_node_data(kwargs, wrap=True)
        return tf.nest.map_structure(_deserialize_keras_tensor, kwargs)

    def process_node(layer, node_data):
        """Deserialize a node.

        Args:
            layer: layer instance.
            node_data: Nested structure of `ListWrapper`.

        Returns:
            Whether the node was processed (i.e. the layer was called on the
            inputs specified by the node data)

        Raises:
            ValueError: In case of improperly formatted `node_data`.
        """
        input_tensors = []
        for input_data in tf.nest.flatten(node_data):
            input_data = input_data.as_list()
            if len(input_data) == 3:
                kwargs = {}
            elif len(input_data) == 4:
                kwargs = input_data[3]
                try:
                    kwargs = _deserialize_keras_tensors(kwargs, created_layers)
                except IndexError:
                    # Happens if keras tensors in kwargs are still unprocessed
                    return False
            else:
                raise ValueError("Improperly formatted model config.")

            if input_data[0] != node_module._CONSTANT_VALUE:
                inbound_layer_name = input_data[0]
                inbound_node_index = input_data[1]
                inbound_tensor_index = input_data[2]
                inbound_layer = created_layers[inbound_layer_name]
                inbound_node_index = get_node_index(
                    inbound_layer, inbound_node_index
                )

                if inbound_node_index is None:
                    return False
                inbound_node = inbound_layer._inbound_nodes[inbound_node_index]
                input_tensors.append(
                    tf.nest.flatten(inbound_node.outputs)[inbound_tensor_index]
                )
            else:
                # We received a constant w/ no Keras history attached,
                # which means it is a constant tensor input.
                # Input is a constant value.
                # Format = [_CONSTANT_VALUE, -1, const_val, kwargs]
                assert input_data[1] == -1
                assert len(input_data) >= 3
                const_val = input_data[2]
                if (
                    isinstance(const_val, tuple)
                    and len(const_val) == 2
                    and const_val[0] == node_module._COMPOSITE_TYPE
                ):
                    # It is a composite tensor.
                    input_tensors.append(json_utils.decode(const_val[1]))
                else:
                    input_tensors.append(const_val)
        input_tensors = tf.nest.pack_sequence_as(node_data, input_tensors)
        # Call layer on its inputs, thus creating the node
        # and building the layer if needed.
        if input_tensors is not None:
            if not layer._preserve_input_structure_in_config:
                input_tensors = base_layer_utils.unnest_if_single_tensor(
                    input_tensors
                )
            output_tensors = layer(input_tensors, **kwargs)

            # Update node index map.
            output_index = tf.nest.flatten(output_tensors)[
                0
            ]._keras_history.node_index
            node_index_map[
                (layer.name, node_count_by_layer[layer])
            ] = output_index
            node_count_by_layer[layer] += 1
        return True

    def process_layer(layer_data):
        """Deserializes a layer, then call it on appropriate inputs.

        Args:
            layer_data: layer config dict.

        Raises:
            ValueError: In case of improperly formatted `layer_data` dict.
        """
        layer_name = layer_data["name"]

        if layer_name in created_layers:
            layer = created_layers[layer_name]
        else:
            # Instantiate layer.
            from keras.layers import deserialize as deserialize_layer

            layer = deserialize_layer(layer_data, custom_objects=custom_objects)
            created_layers[layer_name] = layer

        node_count_by_layer[layer] = int(_should_skip_first_node(layer))

        # Gather layer inputs and convert to `ListWrapper` objects.
        inbound_nodes_data = layer_data["inbound_nodes"]
        inbound_nodes_data = tf_utils.convert_inner_node_data(
            inbound_nodes_data, wrap=True
        )
        for node_data in inbound_nodes_data:
            # We don't process nodes (i.e. make layer calls)
            # on the fly because the inbound node may not yet exist,
            # in case of layer shared at different topological depths
            # (e.g. a model such as A(B(A(B(x)))))
            unprocessed_nodes[layer].append(node_data)

    # First, we create all layers and enqueue nodes to be processed
    for layer_data in config["layers"]:
        process_layer(layer_data)
    # Then we process nodes in order of layer depth.
    # Nodes that cannot yet be processed (if the inbound node
    # does not yet exist) are re-enqueued, and the process
    # is repeated until all nodes are processed.
    while unprocessed_nodes:
        for layer_data in config["layers"]:
            layer = created_layers[layer_data["name"]]
            if layer in unprocessed_nodes:
                layer_nodes = unprocessed_nodes.pop(layer)
                while layer_nodes:
                    node_data = layer_nodes[0]
                    if process_node(layer, node_data):
                        layer_nodes.pop(0)
                    else:
                        # If a node can't be processed, stop processing the
                        # nodes of the current layer to maintain node ordering.
                        unprocessed_nodes[layer] = layer_nodes
                        break

    input_tensors = []
    output_tensors = []

    input_layers = tf_utils.convert_inner_node_data(
        config["input_layers"], wrap=True
    )
    for layer_data in tf.nest.flatten(input_layers):
        layer_name, node_index, tensor_index = layer_data.as_list()
        assert layer_name in created_layers
        layer = created_layers[layer_name]
        node_index = get_node_index(layer, node_index)
        layer_output_tensors = layer._inbound_nodes[node_index].output_tensors
        input_tensors.append(
            tf.nest.flatten(layer_output_tensors)[tensor_index]
        )

    output_layers = tf_utils.convert_inner_node_data(
        config["output_layers"], wrap=True
    )
    for layer_data in tf.nest.flatten(output_layers):
        layer_name, node_index, tensor_index = layer_data.as_list()
        assert layer_name in created_layers
        layer = created_layers[layer_name]
        node_index = get_node_index(layer, node_index)
        layer_output_tensors = layer._inbound_nodes[node_index].output_tensors
        output_tensors.append(
            tf.nest.flatten(layer_output_tensors)[tensor_index]
        )

    input_tensors = tf.nest.pack_sequence_as(input_layers, input_tensors)
    output_tensors = tf.nest.pack_sequence_as(output_layers, output_tensors)
    return input_tensors, output_tensors, created_layers


def get_network_config(network, serialize_layer_fn=None, config=None):
    """Builds the config, which consists of the node graph and serialized layers.

    Args:
      network: A Network object.
      serialize_layer_fn: Function used to serialize layers.
      config: A dict to append more config entries into. If None, start with a
          new dict for the config.

    Returns:
      Config dictionary.
    """
    serialize_layer_fn = (
        serialize_layer_fn or generic_utils.serialize_keras_object
    )
    config = config or {}
    config["name"] = network.name
    node_conversion_map = {}
    for layer in network.layers:
        kept_nodes = 1 if _should_skip_first_node(layer) else 0
        for original_node_index, node in enumerate(layer._inbound_nodes):
            node_key = _make_node_key(layer.name, original_node_index)
            if node_key in network._network_nodes:
                node_conversion_map[node_key] = kept_nodes
                kept_nodes += 1
    layer_configs = []

    with generic_utils.SharedObjectSavingScope():
        for layer in network.layers:  # From the earliest layers on.
            filtered_inbound_nodes = []
            for original_node_index, node in enumerate(layer._inbound_nodes):
                node_key = _make_node_key(layer.name, original_node_index)
                if node_key in network._network_nodes and not node.is_input:
                    # The node is relevant to the model:
                    # add to filtered_inbound_nodes.
                    node_data = node.serialize(
                        _make_node_key, node_conversion_map
                    )
                    filtered_inbound_nodes.append(node_data)

            layer_config = serialize_layer_fn(layer)
            layer_config["name"] = layer.name
            layer_config["inbound_nodes"] = filtered_inbound_nodes
            layer_configs.append(layer_config)
        config["layers"] = layer_configs

    # Gather info about inputs and outputs.
    model_inputs = []
    for i in range(len(network._input_layers)):
        layer, node_index, tensor_index = network._input_coordinates[i]
        node_key = _make_node_key(layer.name, node_index)
        if node_key not in network._network_nodes:
            continue
        new_node_index = node_conversion_map[node_key]
        model_inputs.append(
            tf_utils.ListWrapper([layer.name, new_node_index, tensor_index])
        )
    model_inputs = tf.nest.pack_sequence_as(
        network._nested_inputs, model_inputs
    )
    # Preserve external Keras compat for Models with single input.
    if not tf.nest.is_nested(model_inputs):
        model_inputs = [model_inputs]
    model_inputs = tf_utils.convert_inner_node_data(model_inputs)
    config["input_layers"] = model_inputs

    model_outputs = []
    for i in range(len(network._output_layers)):
        layer, node_index, tensor_index = network._output_coordinates[i]
        node_key = _make_node_key(layer.name, node_index)
        if node_key not in network._network_nodes:
            continue
        new_node_index = node_conversion_map[node_key]
        model_outputs.append(
            tf_utils.ListWrapper([layer.name, new_node_index, tensor_index])
        )
    model_outputs = tf.nest.pack_sequence_as(
        network._nested_outputs, model_outputs
    )
    # Preserve external Keras compat for Models with single output.
    if not tf.nest.is_nested(model_outputs):
        model_outputs = [model_outputs]
    model_outputs = tf_utils.convert_inner_node_data(model_outputs)
    config["output_layers"] = model_outputs
    return config


def shape_with_no_batch_size(x):
    if x.shape.rank is None:
        return None
    shape = x.shape.as_list()
    if shape:
        shape[0] = None
    return shape


class ModuleWrapper(base_layer.Layer):
    """Wrapper for `tf.Module`s to support the Functional and Sequential API."""

    def __init__(self, module, method_name=None, **kwargs):
        """Initializes the wrapper Layer for this module.

        Args:
          module: The `tf.Module` instance to be wrapped.
          method_name: (Optional) str. The name of the method to use as the
            forward pass of the module. If not set, defaults to '__call__' if
            defined, or 'call'.
          **kwargs: Additional keywrod arguments. See `tf.keras.layers.Layer`.

        Raises:
          ValueError: If `method` is not defined on `module`.
        """
        super().__init__(**kwargs)
        if method_name is None:
            if hasattr(module, "__call__"):
                method_name = "__call__"
            elif hasattr(module, "call"):
                method_name = "call"
        if method_name is None or not hasattr(module, method_name):
            raise ValueError(
                "{} is not defined on object {}".format(method_name, module)
            )

        self._module = module
        self._method_name = method_name

        # Check if module.__call__ has a `training` arg or accepts `**kwargs`.
        method = getattr(module, method_name)
        method_arg_spec = tf_inspect.getfullargspec(method)
        self._call_spec.expects_training_arg = (
            "training" in method_arg_spec.args
            or method_arg_spec.varkw is not None
        )
        self._call_spec.expects_mask_arg = (
            "mask" in method_arg_spec.args or method_arg_spec.varkw is not None
        )

    def call(self, *args, **kwargs):
        if "training" in kwargs and not self._expects_training_arg:
            kwargs.pop("training")
        if "mask" in kwargs and not self._expects_mask_arg:
            kwargs.pop("mask")
        return getattr(self._module, self._method_name)(*args, **kwargs)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Small NumPy datasets for debugging/testing.
"""

import sys as _sys

from keras.api.keras.datasets import boston_housing
from keras.api.keras.datasets import cifar10
from keras.api.keras.datasets import cifar100
from keras.api.keras.datasets import fashion_mnist
from keras.api.keras.datasets import imdb
from keras.api.keras.datasets import mnist
from keras.api.keras.datasets import reuters
from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "keras.datasets", public_apis=None, deprecation=True,
      has_lite=False)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Classes and functions implementing Metrics SavedModel serialization."""

import tensorflow.compat.v2 as tf

from keras.saving.saved_model import constants
from keras.saving.saved_model import layer_serialization
from keras.utils import generic_utils


class MetricSavedModelSaver(layer_serialization.LayerSavedModelSaver):
    """Metric serialization."""

    @property
    def object_identifier(self):
        return constants.METRIC_IDENTIFIER

    def _python_properties_internal(self):
        metadata = dict(
            class_name=generic_utils.get_registered_name(type(self.obj)),
            name=self.obj.name,
            dtype=self.obj.dtype,
        )
        metadata.update(layer_serialization.get_serialized(self.obj))
        if self.obj._build_input_shape is not None:
            metadata["build_input_shape"] = self.obj._build_input_shape
        return metadata

    def _get_serialized_attributes_internal(self, unused_serialization_cache):
        return (
            dict(variables=tf.__internal__.tracking.wrap(self.obj.variables)),
            # TODO(b/135550038): save functions to enable saving custom metrics.
            {},
        )
